from aisy_sca.callbacks.Callbacks import *
import importlib


class CallbackControls:

    def __init__(self, dataset, settings):
        self.settings = settings
        self.dataset = dataset
        self.callbacks = []
        self.builtin_callbacks = {}
        self.custom_callbacks_list = {}

    def add_callback(self, callback, name):
        self.callbacks.append(callback)
        self.builtin_callbacks[name] = callback
        return self.callbacks

    def add_custom_callback(self, callback, callback_name):
        self.custom_callbacks_list[callback_name] = callback

    def create_callbacks(self):
        """
            Initialize callbacks according to settings
            Callbacks can be:
            - Visualization callback for input gradients calculations (implemented in aisy_sca package)
            - Early stopping callbacks if user consider early-stopping metrics
            - Custom callbacks from user
        """

        self.callbacks = []
        if self.settings["use_visualization"]:
            """ Create visualization callback """
            self.add_callback(InputGradients(self.dataset, self.settings), "input_gradients_callback")

        if self.settings["use_early_stopping"]:
            """ Create early-stopping (user custom metrics) callbacks """
            self.add_callback(EarlyStoppingCallback(self.dataset, self.settings), "early_stopping_callback")

        if self.settings["use_confusion_matrix"]:
            """ Create confusion matrix callback """
            x = self.dataset.x_validation if self.dataset.x_validation is not None else self.dataset.x_attack
            y = self.dataset.y_validation if self.dataset.y_validation is not None else self.dataset.y_attack
            self.add_callback(ConfusionMatrix(x, y), "confusion_matrix_callback")

        if self.settings["use_custom_callbacks"]:

            """ Create user custom callbacks """
            for custom_callback in self.settings["custom_callbacks"]:
                callback_inverted = custom_callback['class'][::-1]
                module_name = callback_inverted.partition(".")[2][::-1]
                callback_name = callback_inverted.partition(".")[0][::-1]
                module_name = importlib.import_module(module_name)
                custom_callback_class = getattr(module_name, callback_name)
                custom_callback_obj = custom_callback_class(self.dataset, self.settings, custom_callback['parameters'])
                self.add_callback(custom_callback_obj, callback_name)
                self.add_custom_callback(custom_callback_obj, custom_callback['name'])

    def get_callbacks(self):
        return self.callbacks

    def get_builtin_callbacks(self):
        return self.builtin_callbacks

    def get_custom_callbacks(self):
        return self.custom_callbacks_list



from aisy_sca.datasets import Dataset
from tensorflow.keras.utils import to_categorical
from aisy_sca.crypto.sca_aes_create_intermediates import *
import h5py


class DatasetControls:

    def __init__(self, settings, split_training_set_ratio=0.9, apply_znorm=True):
        self.x_profiling = None
        self.x_validation = None
        self.x_attack = None
        self.y_profiling = None
        self.y_validation = None
        self.y_attack = None
        self.plaintext_profiling = None
        self.plaintext_validation = None
        self.plaintext_attack = None
        self.ciphertext_profiling = None
        self.ciphertext_validation = None
        self.ciphertext_attack = None
        self.key_profiling = None
        self.key_validation = None
        self.key_attack = None
        self.settings = settings
        self.split_training_set_ratio = split_training_set_ratio
        self.apply_znorm = apply_znorm

    def read(self):

        if ".h5" not in self.settings["filename"]:
            print("ERROR 10: Dataset format not supported.")
            return

        """
        Read h5 dataset with the following structure and metadata:
        [Profiling_traces/traces]
        [Profiling_traces/metadata][plaintext]
        [Profiling_traces/metadata][key]
        [Profiling_traces/metadata][ciphertext] (if exists)
        [Attack_traces/traces]
        [Attack_traces/metadata][plaintext]
        [Attack_traces/metadata][key]
        [Attack_traces/metadata][ciphertext] (if exists)
        """

        """ Define main settings"""

        nt = self.settings["number_of_profiling_traces"]
        na = self.settings["number_of_attack_traces"]
        fs = self.settings['first_sample']
        ns = self.settings['number_of_samples']

        """ Read file and get traces and metadata """

        in_file = h5py.File(f"{self.settings['datasets_root_folder']}{self.settings['filename']}", "r")
        self.x_profiling = np.array(in_file['Profiling_traces/traces'], dtype=np.float64)[:nt, fs:fs + ns]
        self.x_attack = np.array(in_file['Attack_traces/traces'], dtype=np.float64)[:na, fs:fs + ns]
        self.plaintext_profiling = in_file['Profiling_traces/metadata']['plaintext'][:nt]
        self.plaintext_attack = in_file['Attack_traces/metadata']['plaintext'][:na]
        self.key_profiling = in_file['Profiling_traces/metadata']['key'][:nt]
        self.key_attack = in_file['Attack_traces/metadata']['key'][:na]

        """ if ciphertext is in metadata """
        if "ciphertext" in in_file['Profiling_traces/metadata'].dtype.fields:
            self.ciphertext_profiling = in_file['Profiling_traces/metadata']['ciphertext'][:nt]
            self.ciphertext_attack = in_file['Attack_traces/metadata']['ciphertext'][:na]

        self.create_labels(self.settings)

        """ Split datasets if required """
        if self.settings["split_test_set"]:
            self.split_test_data(na)
        elif self.settings["split_training_set"]:
            self.split_training_data(nt, self.split_training_set_ratio)

        """ Convert to categorical labels """
        self.y_profiling = to_categorical(self.y_profiling, num_classes=self.settings["classes"])
        if self.y_validation is not None:
            self.y_validation = to_categorical(self.y_validation, num_classes=self.settings["classes"])
        self.y_attack = to_categorical(self.y_attack, num_classes=self.settings["classes"])

        if self.apply_znorm:
            self.apply_znorm_dataset()

    def create_labels(self, settings):

        if settings["leakage_model"]["cipher"] == "AES128":
            self.y_profiling = aes_intermediates(self.plaintext_profiling, self.ciphertext_profiling, self.key_profiling,
                                                 settings["leakage_model"])
            self.y_attack = aes_intermediates(self.plaintext_attack, self.ciphertext_attack, self.key_attack,
                                              settings["leakage_model"])
        else:
            print("ERROR: cipher not supported.")
            return

    def split_test_data(self, na):

        self.x_validation = self.x_attack[0: int(na / 2)]
        self.y_validation = self.y_attack[0: int(na / 2)]
        self.x_attack = self.x_attack[int(na / 2): na]
        self.y_attack = self.y_attack[int(na / 2): na]

        self.plaintext_validation = self.plaintext_attack[0: int(na / 2)]
        self.plaintext_attack = self.plaintext_attack[int(na / 2): na]

        if self.ciphertext_attack is not None:
            self.ciphertext_validation = self.ciphertext_attack[0: int(na / 2)]
            self.ciphertext_attack = self.ciphertext_attack[int(na / 2): na]

        self.key_validation = self.key_attack[0: int(na / 2)]
        self.key_attack = self.key_attack[int(na / 2): na]

    def split_training_data(self, nt, ratio):

        nt = int(nt * ratio)

        self.x_validation = self.x_profiling[nt:]
        self.y_validation = self.y_profiling[nt:]
        self.x_profiling = self.x_profiling[:nt]
        self.y_profiling = self.y_profiling[:nt]

        self.plaintext_validation = self.plaintext_profiling[nt:]
        self.plaintext_profiling = self.plaintext_profiling[:nt]

        if self.ciphertext_attack is not None:
            self.ciphertext_validation = self.ciphertext_profiling[nt:]
            self.ciphertext_profiling = self.ciphertext_profiling[:nt]

        self.key_validation = self.key_profiling[nt:]
        self.key_profiling = self.key_profiling[:nt]

    def create_z_score_norm(self, dataset):
        return np.mean(dataset, axis=0), np.std(dataset, axis=0)

    def apply_z_score_norm(self, dataset, mean, std):
        for index in range(len(dataset)):
            dataset[index] = (dataset[index] - mean) / std

    def apply_znorm_dataset(self):

        mean, std = self.create_z_score_norm(self.x_profiling)
        self.apply_z_score_norm(self.x_profiling, mean, std)
        if self.x_validation is not None:
            self.apply_z_score_norm(self.x_validation, mean, std)
        self.apply_z_score_norm(self.x_attack, mean, std)

        self.x_profiling = self.x_profiling.astype('float32')
        if self.x_validation is not None:
            self.x_validation = self.x_validation.astype('float32')
        self.x_attack = self.x_attack.astype('float32')

    def get_traces(self):
        return self.x_profiling, self.x_validation, self.x_attack

    def get_labels(self):
        return self.y_profiling, self.y_validation, self.y_attack

    def get_plaintexts(self):
        return self.plaintext_profiling, self.plaintext_validation, self.plaintext_attack

    def get_ciphertexts(self):
        return self.ciphertext_profiling, self.ciphertext_validation, self.ciphertext_attack

    def get_keys(self):
        return self.key_profiling, self.key_validation, self.key_attack

    def get_dataset(self):
        return Dataset(self.x_profiling, self.y_profiling, self.x_attack, self.y_attack, self.x_validation, self.y_validation,
                       self.plaintext_profiling, self.plaintext_attack, self.ciphertext_profiling, self.ciphertext_attack,
                       self.key_profiling, self.key_attack, self.plaintext_validation, self.ciphertext_validation, self.key_validation)



from aisy_sca.optimization.EnsembleControls import EnsembleControls
from aisy_sca.AnalysisDatabaseControls import AnalysisDatabaseControls
from sklearn.utils import shuffle
import numpy as np
import random


class Ensemble:

    def __init__(self, settings, dataset):
        self.settings = settings
        self.dataset = dataset

    def compute(self, sk_all_models, number_of_best_models, labels_key_hypothesis_attack, db_controls, early_stopping_metric=None,
                n_profiling=None):

        ens_controls = EnsembleControls(self.settings)
        if early_stopping_metric is not None:
            ge_all_validation, ge_all_attack = db_controls.select_all_early_stopping_guessing_entropy_from_analysis(early_stopping_metric,
                                                                                                                    n_profiling=n_profiling)
            # sr_all_validation, sr_all_attack = db_controls.select_all_early_stopping_success_rate_from_analysis(early_stopping_metric,
            #                                                                                                     n_profiling=n_profiling)
        else:
            ge_all_validation, ge_all_attack = db_controls.select_all_guessing_entropy_from_analysis(n_profiling=n_profiling)
            # sr_all_validation, sr_all_attack = db_controls.select_all_success_rate_from_analysis(n_profiling=n_profiling)

        number_of_models = len(ge_all_validation)

        nt_interval = int(self.settings["key_rank_attack_traces"] / self.settings["key_rank_report_interval"])

        list_of_best_models = ens_controls.order_models_per_guessing_entropy(ge_all_validation, number_of_models)

        # ge_best_model_validation = ge_all_validation[list_of_best_models[0]]
        # ge_best_model_attack = ge_all_attack[list_of_best_models[0]]
        # sr_best_model_validation = sr_all_validation[list_of_best_models[0]]
        # sr_best_model_attack = sr_all_attack[list_of_best_models[0]]

        kr_ensemble = np.zeros(nt_interval)
        krs_ensemble = np.zeros((self.settings["key_rank_executions"], nt_interval))
        kr_ensemble_best_models = np.zeros(nt_interval)
        krs_ensemble_best_models = np.zeros((self.settings["key_rank_executions"], nt_interval))

        nt = len(self.dataset.x_attack)

        probabilities_kg_all_traces = np.zeros((number_of_models, nt, 256))

        for model_index in range(number_of_models):

            out_prob_model = sk_all_models[list_of_best_models[model_index]]
            for index in range(self.settings["key_rank_attack_traces"]):
                probabilities_kg_all_traces[model_index][index] = out_prob_model[index][
                    np.asarray([int(leakage[index]) for leakage in labels_key_hypothesis_attack[:]])
                ]
            print(f"Processing Model {list_of_best_models[model_index]} of {number_of_models}")

        random_states_ensembles = []

        for run in range(self.settings["key_rank_executions"]):

            key_p_ensemble = np.zeros(256)
            key_p_ensemble_best_models = np.zeros(256)

            probabilities_kg_all_traces_shuffled = np.zeros((number_of_models, nt, 256))
            # if self.settings["reproducible"]:
            #     random_state = self.ge_sr_random_states_ensemble[run]
            # else:
            #     random_state = random.randint(0, 100000)
            random_state = random.randint(0, 100000)
            random_states_ensembles.append(random_state)
            for model_index in range(number_of_models):
                probabilities_kg_all_traces_shuffled[model_index] = shuffle(probabilities_kg_all_traces[model_index],
                                                                            random_state=random_state)

            kr_count = 0
            for index in range(self.settings["key_rank_attack_traces"]):
                for model_index in range(number_of_models):
                    key_p_ensemble += np.log(probabilities_kg_all_traces_shuffled[model_index][index] + 1e-36)
                for model_index in range(number_of_best_models):
                    key_p_ensemble_best_models += np.log(probabilities_kg_all_traces_shuffled[model_index][index] + 1e-36)

                key_p_ensemble_sorted = np.argsort(key_p_ensemble)[::-1]
                key_p_ensemble_best_models_sorted = np.argsort(key_p_ensemble_best_models)[::-1]

                if (index + 1) % self.settings["key_rank_report_interval"] == 0:
                    kr_position = list(key_p_ensemble_sorted).index(self.settings["good_key"]) + 1
                    kr_ensemble[kr_count] += kr_position
                    krs_ensemble[run][kr_count] = kr_position

                    kr_position = list(key_p_ensemble_best_models_sorted).index(self.settings["good_key"]) + 1
                    kr_ensemble_best_models[kr_count] += kr_position
                    krs_ensemble_best_models[run][kr_count] = kr_position
                    kr_count += 1

            print("Run: {} | GE {} models: {} | GE {} best models: {}".format(run, number_of_models,
                                                                              kr_ensemble[nt_interval - 1] / (run + 1),
                                                                              number_of_best_models,
                                                                              kr_ensemble_best_models[nt_interval - 1] / (run + 1)))

        self.settings["ge_sr_random_states_ensemble"] = random_states_ensembles

        ge_ensemble = kr_ensemble / self.settings["key_rank_executions"]
        ge_ensemble_best_models = kr_ensemble_best_models / self.settings["key_rank_executions"]

        sr_ensemble = np.zeros(nt_interval)
        sr_ensemble_best_models = np.zeros(nt_interval)

        for index in range(nt_interval):
            for run in range(self.settings["key_rank_executions"]):
                sr_ensemble[index] += 1 if krs_ensemble[run][index] == 1 else 0
                sr_ensemble_best_models[index] += 1 if krs_ensemble_best_models[run][index] == 1 else 0

        sr_ensemble = sr_ensemble / self.settings["key_rank_executions"]
        sr_ensemble_best_models = sr_ensemble_best_models / self.settings["key_rank_executions"]

        if early_stopping_metric is not None:
            if n_profiling is not None:
                label_ensemble = f"Ensemble {early_stopping_metric} {n_profiling} traces"
            else:
                label_ensemble = f"Ensemble {early_stopping_metric}"
        else:
            if n_profiling is not None:
                label_ensemble = f"Ensemble {n_profiling} traces"
            else:
                label_ensemble = "Ensemble"

        """ save guessing entropy from ensemble calculation to database"""
        db_controls.save_ensemble_guessing_entropy_results(ge_ensemble, label_ensemble)
        db_controls.save_ensemble_guessing_entropy_results(ge_ensemble_best_models,
                                                           f"{label_ensemble} of Best {number_of_best_models} Models")

        """ save success_rate from ensemble calculation to database"""
        db_controls.save_ensemble_success_rate_results(sr_ensemble, label_ensemble)
        db_controls.save_ensemble_success_rate_results(sr_ensemble_best_models, f"{label_ensemble} of Best {number_of_best_models} Models")



import numpy as np

s_box = np.array([
    0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5, 0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
    0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0, 0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
    0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC, 0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
    0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A, 0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
    0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0, 0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
    0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B, 0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
    0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85, 0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
    0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5, 0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
    0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17, 0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
    0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88, 0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
    0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C, 0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
    0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9, 0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
    0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6, 0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
    0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E, 0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
    0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94, 0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
    0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68, 0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16,
])

inv_s_box = np.array([
    0x52, 0x09, 0x6A, 0xD5, 0x30, 0x36, 0xA5, 0x38, 0xBF, 0x40, 0xA3, 0x9E, 0x81, 0xF3, 0xD7, 0xFB,
    0x7C, 0xE3, 0x39, 0x82, 0x9B, 0x2F, 0xFF, 0x87, 0x34, 0x8E, 0x43, 0x44, 0xC4, 0xDE, 0xE9, 0xCB,
    0x54, 0x7B, 0x94, 0x32, 0xA6, 0xC2, 0x23, 0x3D, 0xEE, 0x4C, 0x95, 0x0B, 0x42, 0xFA, 0xC3, 0x4E,
    0x08, 0x2E, 0xA1, 0x66, 0x28, 0xD9, 0x24, 0xB2, 0x76, 0x5B, 0xA2, 0x49, 0x6D, 0x8B, 0xD1, 0x25,
    0x72, 0xF8, 0xF6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xD4, 0xA4, 0x5C, 0xCC, 0x5D, 0x65, 0xB6, 0x92,
    0x6C, 0x70, 0x48, 0x50, 0xFD, 0xED, 0xB9, 0xDA, 0x5E, 0x15, 0x46, 0x57, 0xA7, 0x8D, 0x9D, 0x84,
    0x90, 0xD8, 0xAB, 0x00, 0x8C, 0xBC, 0xD3, 0x0A, 0xF7, 0xE4, 0x58, 0x05, 0xB8, 0xB3, 0x45, 0x06,
    0xD0, 0x2C, 0x1E, 0x8F, 0xCA, 0x3F, 0x0F, 0x02, 0xC1, 0xAF, 0xBD, 0x03, 0x01, 0x13, 0x8A, 0x6B,
    0x3A, 0x91, 0x11, 0x41, 0x4F, 0x67, 0xDC, 0xEA, 0x97, 0xF2, 0xCF, 0xCE, 0xF0, 0xB4, 0xE6, 0x73,
    0x96, 0xAC, 0x74, 0x22, 0xE7, 0xAD, 0x35, 0x85, 0xE2, 0xF9, 0x37, 0xE8, 0x1C, 0x75, 0xDF, 0x6E,
    0x47, 0xF1, 0x1A, 0x71, 0x1D, 0x29, 0xC5, 0x89, 0x6F, 0xB7, 0x62, 0x0E, 0xAA, 0x18, 0xBE, 0x1B,
    0xFC, 0x56, 0x3E, 0x4B, 0xC6, 0xD2, 0x79, 0x20, 0x9A, 0xDB, 0xC0, 0xFE, 0x78, 0xCD, 0x5A, 0xF4,
    0x1F, 0xDD, 0xA8, 0x33, 0x88, 0x07, 0xC7, 0x31, 0xB1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xEC, 0x5F,
    0x60, 0x51, 0x7F, 0xA9, 0x19, 0xB5, 0x4A, 0x0D, 0x2D, 0xE5, 0x7A, 0x9F, 0x93, 0xC9, 0x9C, 0xEF,
    0xA0, 0xE0, 0x3B, 0x4D, 0xAE, 0x2A, 0xF5, 0xB0, 0xC8, 0xEB, 0xBB, 0x3C, 0x83, 0x53, 0x99, 0x61,
    0x17, 0x2B, 0x04, 0x7E, 0xBA, 0x77, 0xD6, 0x26, 0xE1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0C, 0x7D,
])

shift_row_mask = np.array([0, 5, 10, 15, 4, 9, 14, 3, 8, 13, 2, 7, 12, 1, 6, 11])
inv_shift_row_mask = np.array([0, 13, 10, 7, 4, 1, 14, 11, 8, 5, 2, 15, 12, 9, 6, 3])


def sub_bytes(s):
    return s_box[s]


def inv_sub_bytes(s):
    return inv_s_box[s]


def shift_rows(s):
    s = np.array(s)
    return s[shift_row_mask]


def inv_shift_rows(s):
    s = np.array(s)
    return s[inv_shift_row_mask]


def add_round_key(s, k):
    return [s[i] ^ k[i] for i in range(len(s))]


xtime = lambda a: (((a << 1) ^ 0x1B) & 0xFF) if (a & 0x80) else (a << 1)


def mix_single_column(a):
    t = a[0] ^ a[1] ^ a[2] ^ a[3]
    u = a[0]
    a[0] ^= t ^ xtime(a[0] ^ a[1])
    a[1] ^= t ^ xtime(a[1] ^ a[2])
    a[2] ^= t ^ xtime(a[2] ^ a[3])
    a[3] ^= t ^ xtime(a[3] ^ u)
    return a


def mix_columns(s):
    for i in range(4):
        s[i * 4:i * 4 + 4] = mix_single_column(s[i * 4:i * 4 + 4])
    return s


def inv_mix_columns(s):
    for i in range(4):
        u = xtime(xtime(s[i * 4] ^ s[i * 4 + 2]))
        v = xtime(xtime(s[i * 4 + 1] ^ s[i * 4 + 3]))
        s[i * 4] ^= u
        s[i * 4 + 1] ^= v
        s[i * 4 + 2] ^= u
        s[i * 4 + 3] ^= v
    s = mix_columns(s)
    return s


r_con = (
    0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40,
    0x80, 0x1B, 0x36, 0x6C, 0xD8, 0xAB, 0x4D, 0x9A,
    0x2F, 0x5E, 0xBC, 0x63, 0xC6, 0x97, 0x35, 0x6A,
    0xD4, 0xB3, 0x7D, 0xFA, 0xEF, 0xC5, 0x91, 0x39,
)


def xor_bytes(a, b):
    return [a[i] ^ b[i] for i in range(len(a))]


def expand_key(master_key):
    iteration_count = 0
    for i in range(4, 44):
        word = list(master_key[len(master_key) - 4:])

        if i % 4 == 0:
            word.append(word.pop(0))
            word = [s_box[b] for b in word]
            word[0] ^= r_con[i // 4]

        word = xor_bytes(word, master_key[iteration_count * 4:iteration_count * 4 + 4])
        for w in word:
            master_key.append(w)

        iteration_count += 1

    return [master_key[16 * i: 16 * (i + 1)] for i in range(len(master_key) // 16)]


def encrypt_block(plaintext, key):
    """
    :argument:
        plaintext (numpy array)
        key (numpy array)
    :return:
        ciphertext (numpy array)
    """

    round_keys = expand_key(key)
    state = add_round_key(plaintext, round_keys[0])

    for i in range(1, 10):
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state = add_round_key(state, round_keys[i])

    state = sub_bytes(state)
    state = shift_rows(state)
    state = add_round_key(state, round_keys[-1])
    return state


def decrypt_block(ciphertext, key):
    """
    :argument:
        ciphertext (numpy array)
        key (numpy array)
    :return:
        plaintext (numpy array)
    """

    round_keys = expand_key(key)
    state = add_round_key(ciphertext, round_keys[10])
    state = inv_shift_rows(state)
    state = inv_sub_bytes(state)

    for i in range(9, 0, -1):
        state = add_round_key(state, round_keys[i])
        state = inv_mix_columns(state)
        state = inv_shift_rows(state)
        state = inv_sub_bytes(state)

    return add_round_key(state, round_keys[0])


def get_round_key(key, round):
    return expand_key(key)[round]


def run_test():
    key = ([int(x) for x in bytearray.fromhex("2b7e151628aed2a6abf7158809cf4f3c")])
    print(expand_key(key))
    # key as list of integets

    plaintext = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1])
    print(plaintext)

    ciphertext = encrypt_block(plaintext, key)
    print(ciphertext)

    key = ([int(x) for x in bytearray.fromhex("2b7e151628aed2a6abf7158809cf4f3c")])
    plaintext2 = decrypt_block(ciphertext, key)
    print(plaintext2)

# run_test()



from .Aisy import *
from .callbacks.Callbacks import *
from .callbacks.CallbackControls import *
from aisy_sca.crypto.aes.aes128 import *
from .datasets.Dataset import *
from .datasets.DatasetControls import *
from .metrics.SCAMetrics import *
from .optimization.RandomGridSearch import *
from .optimization.HyperparameterSearch import *
from .optimization.Ensemble import *
from .optimization.KerasMLPRandom import *
from .optimization.KerasCNNRandom import *
from .optimization.KerasMLPGrid import *
from .optimization.KerasCNNGrid import *
from .sca.Profiling import *
from .utils.utils import *



from aisy_sca.crypto.sca_aes_create_intermediates import *
import numpy as np


class LeakageModelControls:

    def __init__(self, settings):
        self.settings = settings

    def set(self, leakage_model="HW", bit=0, byte=0, round=1, round_first=1, round_second=1, cipher="AES128",
            target_state="Sbox", target_state_first="Sbox", target_state_second="Sbox",
            direction="Encryption", attack_direction="input"):

        """
        Function to set the AES Leakage Model in the profiled SCA execution.
        :parameter
            leakage_model: 'HW', 'ID' or 'bit'
            bit: index of target bit (min 0, max 7)
            byte: index of target key byte
            round: index the target round
            round_first: index the first target round
            round_second: index the second target seround
            target_state: 'Sbox', InvSbox', 'AddRoundKey', 'MixColumns', 'InvMixColumns', 'ShiftRows', 'InvShiftRows'
            target_state_first: 'Input', 'Sbox', InvSbox', 'AddRoundKey', 'MixColumns', 'InvMixColumns', 'ShiftRows', 'InvShiftRows',
            'Output'
            target_state_second: 'Input', 'Sbox', InvSbox', 'AddRoundKey', 'MixColumns', 'InvMixColumns', 'ShiftRows', 'InvShiftRows',
            'Output'
            attack_direction: input, output
            direction: 'Encryption', 'Decryption'

        :return
            dictionary containing AES leakage model information:

            self.leakage_model = {
                "leakage_model": leakage_model,
                "bit": bit,
                "byte": byte,
                "round": round,
                "round_first": round_first,
                "round_second": round_second,
                "target_state": target_state,
                "target_state_first": target_state_first,
                "target_state_second": target_state_second,
                "direction": direction,
                "attack_direction": input
            }

        """

        self.settings["leakage_model"] = {
            "leakage_model": leakage_model,
            "bit": bit,
            "byte": byte,
            "round": round,
            "round_first": round_first,  # for Hamming Distance
            "round_second": round_second,  # for Hamming Distance
            "cipher": cipher,
            "target_state": target_state,
            "target_state_first": target_state_first,  # for Hamming Distance
            "target_state_second": target_state_second,  # for Hamming Distance
            "direction": direction,
            "attack_direction": attack_direction
        }

        if self.settings is not None:
            if self.settings["leakage_model"]["leakage_model"] == "HW" or self.settings["leakage_model"]["leakage_model"] == "HD":
                self.settings["classes"] = 9
            elif self.settings["leakage_model"]["leakage_model"] == "ID":
                self.settings["classes"] = 256
            else:
                self.settings["classes"] = 2
        else:
            print("Parameters (param) from target is not selected. Set target before the leakage model.")

        self.settings["leakage_model_is_set"] = True

        return self.settings

    def compute_labels_key_guesses(self, dataset):

        """ Function to compute all intermediate (labels) for all possible key guesses for AES """

        labels_key_guesses_attack_set = np.zeros((256, len(dataset.x_attack)), dtype="int64")
        for key_byte_hypothesis in range(256):
            key_h = bytearray.fromhex(self.settings["key"])
            key_h[self.settings["leakage_model"]["byte"]] = key_byte_hypothesis
            labels_key_guesses_attack_set[key_byte_hypothesis][:] = aes_intermediates(dataset.plaintext_attack,
                                                                                      dataset.ciphertext_attack,
                                                                                      key_h, self.settings["leakage_model"])

        if dataset.x_validation is not None:
            labels_key_guesses_validation_set = np.zeros((256, len(dataset.x_validation)), dtype="int64")
            for key_byte_hypothesis in range(256):
                key_h = bytearray.fromhex(self.settings["key"])
                key_h[self.settings["leakage_model"]["byte"]] = key_byte_hypothesis
                labels_key_guesses_validation_set[key_byte_hypothesis][:] = aes_intermediates(dataset.plaintext_validation,
                                                                                              dataset.ciphertext_validation,
                                                                                              key_h, self.settings["leakage_model"])
        else:
            labels_key_guesses_validation_set = None
        return labels_key_guesses_attack_set, labels_key_guesses_validation_set



from aisy_sca.crypto.aes.aes128 import *


def aes_intermediates_encrypt_from_input(plaintexts, keys, leakage_model):
    if leakage_model["leakage_model"] in ["HD", "HD_ID", "HD_bit"]:
        return get_hd_intermediates_encrypt_input(plaintexts, keys, leakage_model)
    elif leakage_model["leakage_model"] in ["HW", "ID", "bit"]:
        return get_hw_id_intermediates_encrypt_input(plaintexts, keys, leakage_model)
    else:
        print("Leakage model not supported!")
        return None


def aes_intermediates_encrypt_from_output(ciphertexts, keys, leakage_model):
    if leakage_model["leakage_model"] in ["HD", "HD_ID", "HD_bit"]:
        return get_hd_intermediates_encrypt_output(ciphertexts, keys, leakage_model)
    elif leakage_model["leakage_model"] in ["HW", "ID", "bit"]:
        return get_hw_id_intermediates_encrypt_output(ciphertexts, keys, leakage_model)
    else:
        print("Leakage model not supported!")
        return None


def get_sbox_in_xor_sbox_out_encrypt_input(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    state1 = [int(p) ^ int(k) for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]
    state2 = s_box[state1]
    return [int(s1) ^ int(s2) for s1, s2 in zip(np.asarray(state1[:]), np.asarray(state2[:]))]


def get_sbox_in_xor_mixcolumns_out_encrypt_input(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state1 = []
    state2 = []
    for plaintext, key in zip(plaintexts, keys):
        state = add_round_key(plaintext, key)
        state1.append(state)
        state = sub_bytes(state)
        state = shift_rows(state)
        state2.append(mix_columns(state))
    return [int(inv_shift_rows(s1)[leakage_model["byte"]]) ^ int(s2[leakage_model["byte"]]) for s1, s2 in
            zip(np.asarray(state1[:]), np.asarray(state2[:]))]


def get_addroundkey_out_encrypt_input(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [int(p) ^ int(k) for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_sbox_out_encrypt_input(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [s_box[int(p) ^ int(k)] for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_shiftrows_out_encrypt_input(leakage_model, plaintexts, keys):
    plaintext = [row[inv_shift_row_mask[leakage_model["byte"]]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [s_box[int(p) ^ int(k)] for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_mixcolumns_out_encrypt_input(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_mc = []
    for plaintext, key in zip(plaintexts, keys):
        state = add_round_key(plaintext, key)
        state = sub_bytes(state)
        state = shift_rows(state)
        state_mc.append(mix_columns(state))
    return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_mc]


def get_addroundkey_out_encrypt_input_round_2(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_add_round_key = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, key)
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state_add_round_key.append(add_round_key(state, round_keys[1]))  # todo: compute the constant value for rk1
    return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_add_round_key]


def get_sbox_out_encrypt_input_round_2(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_sbox = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, round_keys[0])
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state = add_round_key(state, round_keys[1])  # todo: compute the constant value for rk1
        state_sbox.append(sub_bytes(state))
    return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_sbox]


def get_hd_intermediates_encrypt_input(plaintexts, keys, leakage_model):
    """
            There is support for the following leakage models:
            - HD(SboxIn XOR SboxOut) or SboxIn XOR SboxOut
            - HD(SboxIn XOR MixColumnsOut) or SboxIn XOR MixColumnsOut
            """
    if leakage_model["target_state_first"] == "AddRoundKey" and leakage_model["target_state_second"] == "Sbox" and leakage_model[
        "round_first"] == 1:
        return get_sbox_in_xor_sbox_out_encrypt_input(leakage_model, plaintexts, keys)
    elif leakage_model["target_state_first"] == "AddRoundKey" and leakage_model["target_state_second"] == "MixColumns" and \
            leakage_model["round_first"] == 1:
        return get_sbox_in_xor_mixcolumns_out_encrypt_input(leakage_model, plaintexts, keys)
    else:
        print("Leakage model not supported!")
        return None


def get_hw_id_intermediates_encrypt_input(plaintexts, keys, leakage_model):
    if leakage_model["round"] == 1:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_out_encrypt_input(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "Sbox":
            return get_sbox_out_encrypt_input(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "ShiftRows":
            return get_shiftrows_out_encrypt_input(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "MixColumns":
            return get_mixcolumns_out_encrypt_input(leakage_model, plaintexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    elif leakage_model["round"] == 2:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_out_encrypt_input_round_2(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "Sbox":
            return get_sbox_out_encrypt_input_round_2(leakage_model, plaintexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None


def get_sbox_in_xor_output_encrypt_output(leakage_model, ciphertexts, keys):
    state1 = [row[shift_row_mask[leakage_model["byte"]]] for row in ciphertexts]
    ciphertext = [row[leakage_model["byte"]] for row in ciphertexts]
    key = [row[10][leakage_model["byte"]] for row in keys]
    return [inv_s_box[int(c1) ^ int(k)] ^ int(c2) for c1, c2, k in zip(np.asarray(ciphertext[:]), np.asarray(state1[:]),
                                                                       np.asarray(key[:]))]


def get_sbox_in_xor_sbox_out_encrypt_output(leakage_model, ciphertexts, keys):
    ciphertext = [row[shift_row_mask[leakage_model["byte"]]] for row in ciphertexts]
    key = [row[10][leakage_model["byte"]] for row in keys]
    state1 = [int(c) ^ int(k) for c, k in zip(np.asarray(ciphertext[:]), np.asarray(key[:]))]
    state2 = inv_s_box[state1]
    return [int(s1) ^ int(s2) for s1, s2 in zip(np.asarray(state1[:]), np.asarray(state2[:]))]


def get_addroundkey_in_encrypt_output(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_add_round_key = []
    for ciphertext, key in zip(ciphertexts, keys):
        state_add_round_key.append(add_round_key(ciphertext, key[10]))
    return [int(s[leakage_model["byte"]]) for s in state_add_round_key]


def get_shiftrows_in_encrypt_output(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_shiftrows = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state_shiftrows.append(inv_shift_rows(state))
    return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_shiftrows]


def get_sbox_in_in_encrypt_output(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_sbox = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state = inv_shift_rows(state)
        state_sbox.append(inv_sub_bytes(state))
    return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_sbox]


def get_addroundkey_in_encrypt_output_round_9(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_add_round_key = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state = inv_shift_rows(state)
        state = inv_sub_bytes(state)
        state_add_round_key.append(add_round_key(state, key[9]))  # todo: compute the constant value for rk1
    return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_add_round_key]


def get_mixcolumns_in_encrypt_output_round_9(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(ciphertexts, dtype=np.uint8)
    state_mc = []
    for ciphertext, key in zip(plaintexts, keys):
        state = add_round_key(ciphertext, key[10])
        state = inv_shift_rows(state)
        state = inv_sub_bytes(state)
        state = add_round_key(state, key[9])  # todo: compute the constant value for rk1
        state_mc.append(inv_mix_columns(state))
    return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_mc]


def get_hd_intermediates_encrypt_output(ciphertexts, keys, leakage_model):
    """
        There is support for the following leakage models:
        - HD(SboxIn XOR Ciphertext) or SboxIn XOR Ciphertext
        - HD(SboxIn XOR SboxOut) or SboxIn XOR SboxOut
    """
    if leakage_model["round_first"] == 10:
        if leakage_model["target_state_first"] == "Output" and leakage_model["target_state_second"] == "Sbox":
            return get_sbox_in_xor_output_encrypt_output(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state_first"] == "AddRoundKey" and leakage_model["target_state_second"] == "Sbox":
            return get_sbox_in_xor_sbox_out_encrypt_output(leakage_model, ciphertexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None


def get_hw_id_intermediates_encrypt_output(ciphertexts, keys, leakage_model):
    if leakage_model["round"] == 10:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_in_encrypt_output(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "InvShiftRows":
            return get_shiftrows_in_encrypt_output(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "InvSbox":
            return get_sbox_in_in_encrypt_output(leakage_model, ciphertexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    elif leakage_model["round"] == 9:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_in_encrypt_output_round_9(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "invMixColumns":
            return get_mixcolumns_in_encrypt_output_round_9(leakage_model, ciphertexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None



from tensorflow.keras.optimizers import *
from tensorflow.keras.layers import *
from tensorflow.keras.models import *
import tensorflow as tf
import numpy as np
import random
import importlib


class KerasMLPRandom:

    def __init__(self, settings, hyperparameters, fixed_model=False, loss_function=None):
        self.settings = settings
        self.hyper_parameters = hyperparameters
        self.model_structure = settings["random_search"]["structure"]
        self.fixed_model = fixed_model
        self.loss_function = loss_function

    def mlp_random_search(self):
        if self.fixed_model:
            neurons = self.hyper_parameters["neurons"] if "neurons" in self.hyper_parameters else 50
            layers = self.hyper_parameters["layers"] if "layers" in self.hyper_parameters else 5
            activation = self.hyper_parameters["activation"] if "activation" in self.hyper_parameters else "relu"
            learning_rate = self.hyper_parameters["learning_rate"] if "learning_rate" in self.hyper_parameters else 0.001
            optimizer = self.hyper_parameters["optimizer"] if "optimizer" in self.hyper_parameters else "Adam"
            dropout_rate = self.hyper_parameters["dropout_rate"] if "dropout_rate" in self.hyper_parameters else 0
            random_seed = self.hyper_parameters["seed"] if "seed" in self.hyper_parameters else np.random.randint(1048576)
        else:
            # set default values if not present in hyper_parameters_ranges
            neurons = random.randrange(self.hyper_parameters["neurons"]["min"], self.hyper_parameters["neurons"]["max"] + self.hyper_parameters["neurons"]["step"],
                                       self.hyper_parameters["neurons"]["step"]) if "neurons" in self.hyper_parameters else 50
            layers = random.randrange(self.hyper_parameters["layers"]["min"], self.hyper_parameters["layers"]["max"] + self.hyper_parameters["layers"]["step"],
                                      self.hyper_parameters["layers"]["step"]) if "layers" in self.hyper_parameters else 5
            activation = random.choice(self.hyper_parameters["activation"]) if "activation" in self.hyper_parameters else "relu"
            learning_rate = random.choice(self.hyper_parameters["learning_rate"]) if "learning_rate" in self.hyper_parameters else 0.001
            optimizer = random.choice(self.hyper_parameters["optimizer"]) if "optimizer" in self.hyper_parameters else "Adam"
            if self.model_structure["use_dropout_after_dense_layer"] or self.model_structure["use_dropout_before_dense_layer"]:
                dropout_rate = random.choice(self.hyper_parameters["dropout_rate"])
            else:
                dropout_rate = 0
            random_seed = np.random.randint(1048576)

        hp = {
            'neurons': neurons,
            'layers': layers,
            'activation': activation,
            'learning_rate': learning_rate,
            'optimizer': optimizer,
            'seed': random_seed
        }

        if self.model_structure["use_dropout_after_dense_layer"] or self.model_structure["use_dropout_before_dense_layer"]:
            hp['dropout_rate'] = dropout_rate

        tf.random.set_seed(random_seed)

        model = Sequential()
        if self.model_structure["use_pooling_before_first_dense"]:
            if "pooling_size_0" in self.hyper_parameters and "pooling_stride_0" in self.hyper_parameters and "pooling_type_0" in self.hyper_parameters:
                pooling_sizes_0 = random.randrange(self.hyper_parameters["pooling_size_0"]["min"],
                                                   self.hyper_parameters["pooling_size_0"]["max"] + self.hyper_parameters["pooling_size_0"]["step"],
                                                   self.hyper_parameters["pooling_size_0"]["step"])
                pooling_stride_0 = random.randrange(self.hyper_parameters["pooling_stride_0"]["min"],
                                                    self.hyper_parameters["pooling_stride_0"]["max"] + self.hyper_parameters["pooling_stride_0"]["step"],
                                                    self.hyper_parameters["pooling_stride_0"]["step"])
                pooling_type_0 = random.choice(self.hyper_parameters["pooling_type_0"])
                if pooling_type_0 == "average":
                    model.add(AveragePooling1D(pool_size=pooling_sizes_0, strides=pooling_stride_0),
                              input_shape=(self.settings["number_of_samples"], 1))
                else:
                    model.add(MaxPool1D(pool_size=pooling_sizes_0, strides=pooling_stride_0,
                                        input_shape=(self.settings["number_of_samples"], 1)))
                model.add(Flatten())
                if self.model_structure["use_dropout_before_dense_layer"]:
                    model.add(Dropout(dropout_rate))
                model.add(Dense(neurons, activation=activation, input_shape=(self.settings["number_of_samples"],)))

                hp["pooling_size_0"] = pooling_sizes_0
                hp["pooling_stride_0"] = pooling_stride_0
                hp["pooling_type_0"] = pooling_type_0
        else:
            if self.model_structure["use_dropout_before_dense_layer"]:
                model.add(Dropout(dropout_rate))
            model.add(Dense(neurons, activation=activation, input_shape=(self.settings["number_of_samples"],)))
        if self.model_structure["use_dropout_after_dense_layer"]:
            model.add(Dropout(dropout_rate))
        for layer_index in range(layers - 1):
            if self.model_structure["use_dropout_before_dense_layer"]:
                model.add(Dropout(dropout_rate))
            model.add(Dense(neurons, activation=activation))
            if self.model_structure["use_dropout_after_dense_layer"]:
                model.add(Dropout(dropout_rate))
        if self.model_structure["use_dropout_before_dense_layer"]:
            model.add(Dropout(dropout_rate))
        model.add(Dense(self.settings["classes"], activation='softmax'))
        model.summary()

        if self.loss_function is not None:
            class_inverted = self.loss_function["class"][::-1]
            module_name = class_inverted.partition(".")[2][::-1]
            loss_function_name = class_inverted.partition(".")[0][::-1]
            module_name = importlib.import_module(module_name)
            custom_loss_class = getattr(module_name, loss_function_name)
            loss = custom_loss_class(self.settings, parameters=self.loss_function['parameters'])
            model.compile(loss=loss, optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        else:
            model.compile(loss='categorical_crossentropy', optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        return model, hp

    def get_optimizer(self, optimizer, learning_rate):
        if optimizer == "Adam":
            return Adam(lr=learning_rate)
        elif optimizer == "RMSprop":
            return RMSprop(lr=learning_rate)
        elif optimizer == "Adadelta":
            return Adadelta(lr=learning_rate)
        elif optimizer == "Adagrad":
            return Adagrad(lr=learning_rate)
        elif optimizer == "SGD":
            return SGD(lr=learning_rate, momentum=0.9, nesterov=True)
        else:
            return Adam(lr=learning_rate)



from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as backend
import tensorflow as tf
import numpy as np
import importlib
from sklearn.metrics import confusion_matrix
from termcolor import colored
import os


class EarlyStoppingCallback(Callback):
    def __init__(self, dataset, settings):
        super().__init__()
        self.dataset = dataset
        self.settings = settings
        self.metrics = settings["early_stopping"]["metrics"]
        self.timestamp = settings["timestamp"]

        self.metric_results = {}
        for metric in self.metrics:
            self.metric_results[metric] = []

    def on_epoch_end(self, epoch, logs={}):

        if self.metrics is not None:

            for metric in self.metrics:
                script = importlib.import_module(self.metrics[metric]["class"])
                metric_value = script.run(self.dataset, self.settings, self.model, self.metrics[metric]["parameters"])
                self.metric_results[metric].append(metric_value)

                if epoch > 0:
                    ref_value_epoch_list = None
                    ref_metric = None
                    if self.metrics[metric]["direction"] == "max":
                        if isinstance(metric_value, list):
                            ref_value_epoch_list = np.zeros(len(self.metric_results[metric][epoch]))
                            for index, current_epoch_value in enumerate(self.metric_results[metric][epoch]):
                                all_values_metric = []
                                for e in range(epoch):
                                    all_values_metric.append(self.metric_results[metric][e][index])
                                ref_value_epoch_list[index] = np.max(all_values_metric)
                        else:
                            ref_metric = np.max(self.metric_results[metric][0:epoch])
                    else:
                        if isinstance(metric_value, list):
                            ref_value_epoch_list = np.zeros(len(self.metric_results[metric][epoch]))
                            for index, current_epoch_value in enumerate(self.metric_results[metric][epoch]):
                                all_values_metric = []
                                for e in range(epoch):
                                    all_values_metric.append(self.metric_results[metric][e][index])
                                ref_value_epoch_list[index] = np.min(all_values_metric)
                        else:
                            ref_metric = np.min(self.metric_results[metric][0:epoch])

                    if isinstance(metric_value, list):
                        for index, current_epoch_value in enumerate(self.metric_results[metric][epoch]):
                            if (self.metric_results[metric][epoch][index] > ref_value_epoch_list[index] and self.metrics[metric][
                                "direction"] == "max") or (
                                    self.metric_results[metric][epoch][index] < ref_value_epoch_list[index] and self.metrics[metric][
                                "direction"] == "min"):
                                model_name = f"{self.settings['resources_root_folder']}models/best_model_{metric}_{self.settings['timestamp']}_{index}.h5"
                                os.remove(model_name)
                                self.model.save_weights(model_name)
                                print(
                                    colored(f"\nModel saved {metric} = {self.metric_results[metric][epoch][index]} at epoch {epoch + 1}\n",
                                            "blue"))
                    else:
                        if (self.metric_results[metric][epoch] > ref_metric and self.metrics[metric]["direction"] == "max") or (
                                self.metric_results[metric][epoch] < ref_metric and self.metrics[metric]["direction"] == "min"):
                            model_name = f"{self.settings['resources_root_folder']}models/best_model_{metric}_{self.settings['timestamp']}.h5"
                            os.remove(model_name)
                            self.model.save_weights(model_name)
                            print(colored(f"\nModel saved {metric} = {self.metric_results[metric][epoch]} at epoch {epoch + 1}\n", "blue"))
                else:
                    if isinstance(metric_value, list):
                        for index, current_epoch_value in enumerate(self.metric_results[metric][epoch]):
                            model_name = f"{self.settings['resources_root_folder']}models/best_model_{metric}_{self.settings['timestamp']}_{index}.h5"
                            self.model.save_weights(model_name)
                            print(
                                colored(f"\nModel saved {metric} = {self.metric_results[metric][epoch][index]} at epoch {epoch + 1}\n",
                                        "blue"))
                    else:
                        model_name = f"{self.settings['resources_root_folder']}models/best_model_{metric}_{self.settings['timestamp']}.h5"
                        self.model.save_weights(model_name)
                        print(colored(f"\nModel saved {metric} = {self.metric_results[metric][epoch]} at epoch {epoch + 1}\n", "blue"))

    def get_metric_results(self):
        return self.metric_results


class InputGradients(Callback):
    def __init__(self, dataset, settings):
        super().__init__()
        self.current_epoch = 0
        self.x = dataset.x_profiling[:settings["visualization"]]
        self.y = dataset.y_profiling[:settings["visualization"]]
        self.number_of_epochs = settings["epochs"]
        self.number_of_samples = settings["number_of_samples"]
        self.gradients = np.zeros((settings["epochs"], settings["number_of_samples"]))
        self.gradients_sum = np.zeros(settings["number_of_samples"])

    def on_epoch_end(self, epoch, logs=None):
        input_trace = tf.Variable(self.x)

        with tf.GradientTape() as tape:
            tape.watch(input_trace)
            pred = self.model(input_trace)
            loss = tf.keras.losses.categorical_crossentropy(self.y, pred)

        grad = tape.gradient(loss, input_trace)

        input_gradients = np.zeros(self.number_of_samples)
        for i in range(len(self.x)):
            input_gradients += grad[i].numpy().reshape(self.number_of_samples)

        self.gradients[epoch] = input_gradients
        if np.max(self.gradients[epoch]) != 0:
            self.gradients_sum += np.abs(self.gradients[epoch] / np.max(self.gradients[epoch]))
        else:
            self.gradients_sum += np.abs(self.gradients[epoch])

        backend.clear_session()

    def input_gradients_sum(self):
        return np.abs(self.gradients_sum)

    def input_gradients_epochs(self):
        for e in range(self.number_of_epochs):
            if np.max(self.gradients[e]) != 0:
                self.gradients[e] = np.abs(self.gradients[e] / np.max(self.gradients[e]))
            else:
                self.gradients[e] = np.abs(self.gradients[e])
        return self.gradients


class ConfusionMatrix(Callback):
    def __init__(self, x, y):
        super().__init__()
        self.current_epoch = 0
        self.x = x
        self.y_true = np.argmax(y, axis=1)
        self.cm = None

    def on_epoch_end(self, epoch, logs=None):
        Y_pred = self.model.predict(self.x)
        y_pred = np.argmax(Y_pred, axis=1)
        self.cm = confusion_matrix(y_true=self.y_true, y_pred=y_pred)

    def get_confusion_matrix(self):
        return self.cm



import os


class AnalysisControls:

    def __init__(self, settings, models):
        self.settings = settings
        self.models = models

    def update_settings(self, random_search, grid_search, visualization, early_stopping, data_augmentation, callbacks, ensemble,
                        confusion_matrix, profiling_analyzer):

        self.settings["grid_search"] = grid_search
        self.settings["random_search"] = random_search
        self.settings["ensemble"] = ensemble
        self.settings["early_stopping"] = early_stopping
        self.settings["custom_callbacks"] = callbacks
        self.settings["confusion_matrix"] = confusion_matrix
        self.settings["use_grid_search"] = True if grid_search is not None else False
        self.settings["use_random_search"] = True if random_search is not None else False
        self.settings["use_profiling_analyzer"] = True if profiling_analyzer is not None else False

        if visualization is not None:
            self.settings["use_visualization"] = True
            self.settings["visualization"] = visualization[0]
        else:
            self.settings["use_visualization"] = False

        if data_augmentation is not None:
            self.settings["use_data_augmentation"] = True
            self.settings["data_augmentation"] = [data_augmentation[0].__name__, data_augmentation[1]]
        else:
            self.settings["use_data_augmentation"] = False

        if callbacks is not None:
            self.settings["use_custom_callbacks"] = True
        else:
            self.settings["use_custom_callbacks"] = False

        if early_stopping is not None:
            self.settings["use_early_stopping"] = True
            self.settings["early_stopping"] = early_stopping
        else:
            self.settings["use_early_stopping"] = False

        self.settings["use_ensemble"] = True if ensemble is not None else False
        self.settings["use_confusion_matrix"] = confusion_matrix
        self.settings["split_test_set"] = self.settings["use_ensemble"] or self.settings["use_early_stopping"] or self.settings[
            "use_grid_search"] or self.settings["use_random_search"]
        self.settings["split_training_set"] = False

        if self.settings["use_grid_search"] or self.settings["use_random_search"]:
            self.settings["hyperparameter_search"] = True
        else:
            self.settings["hyperparameter_search"] = False

        """ set default value """
        self.settings["profiling_analyzer_steps"] = [self.settings["number_of_profiling_traces"]]

        """ or set the list of steps in case profiling analyzer is considered"""
        if self.settings["use_profiling_analyzer"]:
            if isinstance(profiling_analyzer["steps"], list):
                self.settings["profiling_analyzer_steps"] = profiling_analyzer["steps"]
            else:
                self.settings["profiling_analyzer_steps"] = []
                for step in range(profiling_analyzer["steps"]):
                    n_profiling_traces = int(self.settings["number_of_profiling_traces"] / profiling_analyzer["steps"])
                    n_profiling_traces = n_profiling_traces * (step + 1)
                    self.settings["profiling_analyzer_steps"].append(n_profiling_traces)

    def check_errors(self):

        if self.settings["database_root_folder"] is None:
            print("ERROR: database root folder not defined. Use method 'set_database_root_folder' to define your database root folder.")
            return

        if self.settings["datasets_root_folder"] is None:
            print("ERROR: dataset root folder not defined. Use method 'set_dataset_root_folder' to define your dataset root folder.")
            return

        if self.settings["resources_root_folder"] is None:
            print("ERROR: resources root folder not defined. Use method 'set_resources_root_folder' to define your resources root folder.")
            return

        for folder in ["databases", "figures", "models", "npz"]:
            dir_resources_id = "{}/{}/".format(self.settings["resources_root_folder"], folder)
            if not os.path.exists(dir_resources_id):
                os.makedirs(dir_resources_id)

        if len(self.models) == 0 and self.settings["grid_search"] is None and self.settings["random_search"] is None:
            print("ERROR 1: neural network model is not defined.")
            return

        if "filename" not in self.settings.keys():
            print("ERROR 2: Dataset 'filename' not specified. Please use set_filename method to specify it.")
            return

        if "key" not in self.settings.keys():
            print("ERROR 3: Dataset 'key' not specified. Please use set_key method to specify it.")
            return

        if "first_sample" not in self.settings.keys():
            print("ERROR 4 : Dataset 'first_sample' not specified. Please use set_first_sample method to specify it.")
            return

        if "number_of_samples" not in self.settings.keys():
            print("ERROR 5: Dataset 'number_of_samples' not specified. Please use set_number_of_samples method to specify it.")
            return

        if "number_of_profiling_traces" not in self.settings.keys():
            print(
                "ERROR 6: Dataset 'number_of_profiling_traces' not specified. Please use set_number_of_profiling_traces method to specify it.")
            return

        if "number_of_attack_traces" not in self.settings.keys():
            print("ERROR 7: Dataset 'number_of_attack_traces' not specified. Please use set_number_of_attack_traces method to specify it.")
            return

        if self.settings["use_ensemble"]:
            if self.settings["number_of_attack_traces"] == self.settings["key_rank_attack_traces"]:
                print("ERROR 8: ensemble feature requires the 'number_of_attack_traces' >= 2 x key_rank_attack_traces.")
                return

        if self.settings["use_grid_search"]:
            if self.settings["use_ensemble"] and self.settings["grid_search"]["stop_condition"]:
                print("ERROR 11: when grid search has a stop condition, ensembles can't be applied.")
                return
            if self.settings["number_of_attack_traces"] == self.settings["key_rank_attack_traces"]:
                print("ERROR 14: grid search feature requires the 'number_of_attack_traces' >= 2 x key_rank_attack_traces.")
                return

        if self.settings["use_random_search"]:
            if self.settings["use_ensemble"]:
                if self.settings["ensemble"][0] > self.settings["random_search"]["max_trials"]:
                    print("ERROR 12: number of models for ensembles can't be larger than maximum number of trials in random search.")
                    return
            if self.settings["use_ensemble"] and self.settings["random_search"]["stop_condition"]:
                print("ERROR 13: when random search has a stop condition, ensembles can't be applied.")
                return
            if self.settings["number_of_attack_traces"] == self.settings["key_rank_attack_traces"]:
                print("ERROR 15: random search feature requires the 'number_of_attack_traces' >= 2 x key_rank_attack_traces.")
                return

        if self.settings["use_early_stopping"]:
            if self.settings["number_of_attack_traces"] == self.settings["key_rank_attack_traces"]:
                print("ERROR 9: early stopping feature requires the 'number_of_attack_traces' >= 2 x key_rank_attack_traces.")
                return



import tensorflow.keras.backend as backend
import numpy as np
import json
import time
import os


class Utils:

    def keras_model_as_string(self, model, model_name):
        model_json = json.loads(model.to_json())

        file_contents = f"def {model_name}(classes, number_of_samples):\n"
        for index, layers in enumerate(model_json["config"]["layers"]):
            if index == 0:
                file_contents += "    model = Sequential(name='{}')\n".format(model_json["config"]["name"])
            if "Conv" in layers["class_name"]:
                input_layer = ", input_shape=({}, 1)".format(layers["config"]["batch_input_shape"][1]) if "batch_input_shape" in layers[
                    "config"] else ""
                file_contents += "    model.add(Conv1D(filters={}, kernel_size={}, strides={}, activation='{}', padding='{}', " \
                                 "kernel_initializer='{}', bias_initializer='{}'{}))\n".format(
                    layers["config"]["filters"], layers["config"]["kernel_size"], layers["config"]["strides"],
                    layers["config"]["activation"], layers["config"]["padding"], layers["config"]["kernel_initializer"]["class_name"],
                    layers["config"]["bias_initializer"]["class_name"], input_layer)
            elif "AveragePooling" in layers["class_name"]:
                file_contents += "    model.add(AveragePooling1D(pool_size={}, strides={}))\n".format(layers["config"]["pool_size"],
                                                                                                      layers["config"]["strides"])
            elif "MaxPool" in layers["class_name"]:
                file_contents += "    model.add(MaxPool1D(pool_size={}, strides={}))\n".format(layers["config"]["pool_size"],
                                                                                               layers["config"]["strides"])
            elif "Norm" in layers["class_name"]:
                file_contents += "    model.add(BatchNormalization())\n"
            elif "Flatten" in layers["class_name"]:
                file_contents += "    model.add(Flatten())\n"
            elif "ActivityRegularization" in layers["class_name"]:
                file_contents += "    model.add(ActivityRegularization(l1={}, l2={}))\n".format(layers["config"]["l1"],
                                                                                                layers["config"]["l2"])
            elif "Activation" in layers["class_name"]:
                file_contents += "    model.add(Activation(activation={})\n".format(layers["config"]["activation"])
            elif "Dropout" in layers["class_name"]:
                file_contents += "    model.add(Dropout({}))\n".format(layers["config"]["rate"])
            elif "Dense" in layers["class_name"]:
                input_layer = ", input_shape=({},)".format(layers["config"]["batch_input_shape"][1]) if "batch_input_shape" in layers[
                    "config"] else ""
                file_contents += "    model.add(Dense({}, activation='{}', kernel_initializer='{}', bias_initializer='{}'{}))\n".format(
                    layers["config"]["units"], layers["config"]["activation"], layers["config"]["kernel_initializer"]["class_name"],
                    layers["config"]["bias_initializer"]["class_name"], input_layer)
        file_contents += "    model.summary()\n"
        file_contents += "    optimizer = {}(lr={})\n".format(model.optimizer.__class__.__name__, backend.eval(model.optimizer.lr), 2)
        metric_names = []
        for metric in model.metrics_names:
            if "loss" not in metric:
                metric_names.append(metric)
        file_contents += "    model.compile(loss='{}', optimizer=optimizer, metrics={})\n".format(model.loss, metric_names)
        file_contents += "    return model"
        return file_contents

    def get_hyperparameters_from_model(self, model):
        model_json = json.loads(model.to_json())

        conv_layers = 0
        kernels = []
        strides = []
        filters = []
        pooling_types = []
        pooling_sizes = []
        pooling_strides = []
        neurons = 0
        dense_layers = 0
        activation = "relu"
        learning_rate = float(backend.eval(model.optimizer.lr))
        optimizer = str(model.optimizer.__class__.__name__)
        dropout_rate = 0
        layers_structure = []

        for index, layers in enumerate(model_json["config"]["layers"]):
            layers_structure.append(layers["class_name"])
            if "Conv" in layers["class_name"]:
                conv_layers += 1
                filters.append(int(layers["config"]["filters"]))
                strides.append(int(layers["config"]["strides"][0]))
                kernels.append(int(layers["config"]["kernel_size"][0]))
                activation = str(layers["config"]["activation"])
            elif "AveragePooling" in layers["class_name"]:
                pooling_types.append("Average")
                pooling_sizes.append(int(layers["config"]["pool_size"][0]))
                pooling_strides.append(int(layers["config"]["strides"][0]))
            elif "MaxPool" in layers["class_name"]:
                pooling_types.append("Max")
                pooling_sizes.append(int(layers["config"]["pool_size"][0]))
                pooling_strides.append(int(layers["config"]["strides"][0]))
            elif "Dropout" in layers["class_name"]:
                dropout_rate = layers["config"]["rate"]
            elif "Dense" in layers["class_name"]:
                if index < len(model_json["config"]["layers"]) - 1:
                    dense_layers += 1
                    neurons = int(layers["config"]["units"])
                    activation = str(layers["config"]["activation"])

        hp_list = {
            "neurons": neurons,
            "layers": dense_layers,
            "activation": activation,
            "learning_rate": float(learning_rate),
            "dropout_rate": dropout_rate,
            "optimizer": optimizer,
            "layers_structure": layers_structure
        }
        if conv_layers > 0:
            hp_list["conv_layers"] = conv_layers
            hp_list["kernels"] = kernels
            hp_list["strides"] = strides
            hp_list["filters"] = filters
            hp_list["pooling_types"] = pooling_types
            hp_list["pooling_sizes"] = pooling_sizes
            hp_list["pooling_strides"] = pooling_strides
        hp_list["parameters"] = model.count_params()

        return hp_list

    def adapt_folder_path(self, path):
        if "/" in path and path[len(path) - 1] != "/":
            path += "/"
        if "\\" in path and path[len(path) - 1] != "\\":
            path += "\\"
        return path

    def get_variable_name(self, var):
        return [tpl[0] for tpl in filter(lambda x: var is x[1], globals().items())]



from aisy_sca.optimization.KerasCNNRandom import KerasCNNRandom
from aisy_sca.optimization.KerasMLPRandom import KerasMLPRandom
from aisy_sca.optimization.KerasCNNGrid import KerasCNNGrid
from aisy_sca.optimization.KerasMLPGrid import KerasMLPGrid
from aisy_sca.metrics.SCAMetrics import *
from termcolor import colored
import numpy as np
import os
import json
import random


class HyperparameterSearch:

    def __init__(self, settings, dataset, random_search_definitions, labels_key_guesses_attack_set, labels_key_guesses_validation_set,
                 analysis_db_controls, random_states=None):
        self.settings = settings

        """ set dataset structure: 
                - profiling, validation and attack traces
                - profiling, validation and attack labels
                - profiling, validation and attack plaintexts
                - profiling, validation and attack ciphertexts
                - profiling, validation and attack keys
                """
        self.dataset = dataset

        """ random search definitions """
        self.random_search_definitions = random_search_definitions
        self.analysis_db_controls = analysis_db_controls

        """ Labels for key guesses """
        self.labels_key_guesses_attack_set = labels_key_guesses_attack_set
        self.labels_key_guesses_validation_set = labels_key_guesses_validation_set

        """ Metrics """
        self.sk_attack = []
        self.sk_attack_pa = {}
        self.sk_attack_es = {}
        self.ge_search = []
        self.sr_search = []
        self.loss_search = []
        self.acc_search = []
        self.ge_search_pa = {}
        self.sr_search_pa = {}
        self.loss_search_pa = {}
        self.acc_search_pa = {}
        self.ge_search_early_stopping = {}
        self.sr_search_early_stopping = {}
        self.best_model_index = 0
        self.best_model_index_pa = {}

        """ List of searched hyper-parameters combinations """
        self.hp_searches = []

        """ custom loss """
        self.loss_function = None

        self.random_states = random_states

    def set_metrics(self):

        if self.settings["profiling_analyzer_steps"] is not None:
            for n_profiling_traces in self.settings["profiling_analyzer_steps"]:
                self.ge_search_pa[f"{n_profiling_traces}"] = []
                self.sr_search_pa[f"{n_profiling_traces}"] = []
                self.loss_search_pa[f"{n_profiling_traces}"] = []
                self.acc_search_pa[f"{n_profiling_traces}"] = []
                self.best_model_index_pa[f"{n_profiling_traces}"] = 0
                self.sk_attack_pa[f"{n_profiling_traces}"] = []

    def get_sk_attack(self):
        return self.sk_attack

    def get_sk_attack_pa(self):
        return self.sk_attack_pa

    def get_sk_attack_es(self):
        return self.sk_attack_es

    def update_epochs_and_batch_size(self, hp, hyperparameters):

        hp_values = self.random_search_definitions["hyper_parameters_search"]

        """ check if user requested search for epochs """
        if "epochs" in hp_values:
            self.settings["epochs"] = hyperparameters["epochs"] if self.random_states is not None else random.randrange(
                hp_values["epochs"]["min"],
                hp_values["epochs"]["max"] + hp_values["epochs"]["step"],
                hp_values["epochs"]["step"])
        hp["epochs"] = self.settings["epochs"]

        """ check if user requested search for batch-size """
        if "batch_size" in hp_values:
            self.settings["batch_size"] = hyperparameters["batch_size"] if self.random_states is not None else random.randrange(
                hp_values["batch_size"]["min"],
                hp_values["batch_size"]["max"] +
                hp_values["batch_size"]["step"],
                hp_values["batch_size"]["step"])
        hp["batch_size"] = self.settings["batch_size"]

    def load_model(self, model, setting, idx=None):

        """ Load weights from saved early stopping model from 'resource/models' directory"""

        if idx is None:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}.h5"
        else:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}_{idx}.h5"
        model.load_weights(model_name)

    def delete_model(self, setting, idx=None):

        """ Load model from 'resource/models' directory """

        if idx is None:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}.h5"
        else:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}_{idx}.h5"
        os.remove(model_name)

    def get_early_stopping_metrics(self, early_stopping_metric_results):
        early_stopping_metric_names = []
        for early_stopping_metric in self.settings["early_stopping"]["metrics"]:
            if isinstance(early_stopping_metric_results[early_stopping_metric][0], list):
                for idx in range(len(early_stopping_metric_results[early_stopping_metric][0])):
                    early_stopping_metric_names.append(f"{early_stopping_metric}_{idx}")
            else:
                early_stopping_metric_names.append(early_stopping_metric)
        return early_stopping_metric_names

    def get_early_stopping_values_epochs(self, early_stopping_metric_results):
        early_stopping_values_epochs = {}
        for early_stopping_metric in self.settings["early_stopping"]["metrics"]:
            if isinstance(early_stopping_metric_results[early_stopping_metric][0], list):
                for idx in range(len(early_stopping_metric_results[early_stopping_metric][0])):
                    early_stopping_values_epochs[f"{early_stopping_metric}_{idx}"] = np.array(
                        early_stopping_metric_results[early_stopping_metric])[:, idx]
            else:
                early_stopping_values_epochs[f"{early_stopping_metric}"] = early_stopping_metric_results[early_stopping_metric]
        return early_stopping_values_epochs

    def generate_random_model(self, hp_values):

        """ Generate a random model from hyper-parameters ranges or recreate the model from reproducible settings"""

        if self.random_search_definitions["neural_network"] == "cnn":
            model, hp = KerasCNNRandom(self.settings, hp_values, fixed_model=False if self.random_states is None else True,
                                       loss_function=self.loss_function).cnn_random_search()
        else:
            model, hp = KerasMLPRandom(self.settings, hp_values, fixed_model=False if self.random_states is None else True,
                                       loss_function=self.loss_function).mlp_random_search()
        self.update_epochs_and_batch_size(hp, hp_values)
        return model, hp

    def generate_grid_model(self, hp_values):

        """ Generate a random model from hyper-parameters ranges or recreate the model from reproducible settings"""

        if self.random_search_definitions["neural_network"] == "cnn":
            model, hp = KerasCNNGrid(self.settings, hp_values, fixed_model=False if self.random_states is None else True,
                                     loss_function=self.loss_function).cnn_grid_search()
        else:
            model, hp = KerasMLPGrid(self.settings, hp_values, fixed_model=False if self.random_states is None else True,
                                     loss_function=self.loss_function).mlp_grid_search()
        self.update_epochs_and_batch_size(hp, hp_values)
        return model, hp

    def set_hyper_parameters_list(self, hp, model):
        hyper_parameters_list = {
            "parameters": model.count_params()
        }
        for key, value in hp.items():
            hyper_parameters_list[key] = value
        return hyper_parameters_list

    def update_hyper_parameters_list(self, hp, ge, sr):
        hp["Val GE"] = ge
        hp["Val SR"] = sr
        return hp

    def check_stop_condition(self, search_index, hp):

        """ Check if stopping condition is satisfied """

        search_metric = self.random_search_definitions["metric"]
        stop_value = self.random_search_definitions["stop_value"]
        if self.random_search_definitions["stop_condition"]:
            if search_metric == "guessing_entropy" and self.ge_search[search_index] <= stop_value:
                self.best_model_index = search_index
            elif search_metric == "success_rate" and self.sr_search[search_index] >= stop_value:
                self.best_model_index = search_index
            elif search_metric == "loss" and self.loss_search[search_index] <= stop_value:
                self.best_model_index = search_index
            elif search_metric == "accuracy" and self.acc_search[search_index] >= stop_value:
                self.best_model_index = search_index
            if self.best_model_index == search_index:
                print(colored(f"\nBest Model: {json.dumps(hp, sort_keys=True, indent=4)}", "green"))
                return True
            return False
        return False

    def get_best_model(self, n_profiling_traces=None):

        """ Get search index of best model so far """

        search_metric = self.random_search_definitions["metric"]

        if n_profiling_traces is None:
            if search_metric == "guessing_entropy":
                self.best_model_index = np.argmin(self.ge_search)
            elif search_metric == "success_rate":
                self.best_model_index = np.argmax(self.sr_search)
            elif search_metric == "loss":
                self.best_model_index = np.argmin(self.loss_search)
            elif search_metric == "accuracy":
                self.best_model_index = np.argmax(self.acc_search)
            else:
                self.best_model_index = np.argmin(self.ge_search)

            return self.best_model_index
        else:
            if search_metric == "guessing_entropy":
                self.best_model_index_pa[f"{n_profiling_traces}"] = int(np.argmin(self.ge_search_pa[f"{n_profiling_traces}"]))
            elif search_metric == "success_rate":
                self.best_model_index_pa[f"{n_profiling_traces}"] = int(np.argmax(self.sr_search_pa[f"{n_profiling_traces}"]))
            elif search_metric == "loss":
                self.best_model_index_pa[f"{n_profiling_traces}"] = int(np.argmin(self.loss_search_pa[f"{n_profiling_traces}"]))
            elif search_metric == "accuracy":
                self.best_model_index_pa[f"{n_profiling_traces}"] = int(np.argmax(self.acc_search_pa[f"{n_profiling_traces}"]))
            else:
                self.best_model_index_pa[f"{n_profiling_traces}"] = int(np.argmin(self.ge_search_pa[f"{n_profiling_traces}"]))

            return self.best_model_index_pa

    def display_best_model(self):

        """  Display best model so far """
        if self.settings["use_profiling_analyzer"]:
            for step in self.settings["profiling_analyzer_steps"]:
                self.get_best_model(step)

                print("=========================================")
                print(f" Best Model for {step} profiling traces ")
                print("=========================================")
                print(
                    colored(
                        "\nBest Model: {}".format(
                            json.dumps(self.hp_searches[self.best_model_index_pa[f"{step}"]], sort_keys=True, indent=4)),
                        "green"))
        else:
            self.get_best_model()
            print(
                colored("\nBest Model: {}".format(json.dumps(self.hp_searches[self.best_model_index], sort_keys=True, indent=4)), "green"))

    def save_best_model(self, model, search_index, n_profiling_traces=None):

        """ Save best model so far in h5 """

        if self.best_model_index == search_index or search_index == 0:
            search_metric = self.random_search_definitions["metric"]
            folder = self.settings["resources_root_folder"]
            analysis_id = self.settings["analysis_id"]
            if n_profiling_traces is None:
                model.save_weights(f"{folder}models/best_model_random_search_{search_metric}_{analysis_id}.h5")
            else:
                model.save_weights(
                    f"{folder}models/best_model_random_search_{search_metric}_{n_profiling_traces}traces_id_{analysis_id}.h5")

    def update_sk(self, sk, n_profiling_traces):
        if self.settings["use_ensemble"]:
            if self.settings["use_profiling_analyzer"]:
                self.sk_attack_pa[f"{n_profiling_traces}"].append(sk)
            else:
                self.sk_attack.append(sk)

    def update_sk_es(self, sk, es_metric, n_profiling_traces):
        if self.settings["use_ensemble"]:
            if self.settings["use_profiling_analyzer"]:
                self.sk_attack_es[f"{es_metric}"][f"{n_profiling_traces}"].append(sk)
            else:
                if self.settings["use_ensemble"]:
                    self.sk_attack_es[f"{es_metric}"].append(sk)

    def update_ge_and_sr(self, ge, sr, n_profiling_traces):
        if self.settings["use_profiling_analyzer"]:
            self.ge_search_pa[f"{n_profiling_traces}"].append(ge[-1])
            self.sr_search_pa[f"{n_profiling_traces}"].append(sr[-1])
        else:
            self.ge_search.append(ge[-1])
            self.sr_search.append(sr[-1])

    def initialize_early_stopping_ge_and_sr(self, es_metric, steps):
        if self.settings["use_profiling_analyzer"]:
            if len(self.ge_search_early_stopping) == 0:
                self.ge_search_early_stopping[f"{es_metric}"] = {}
                self.sr_search_early_stopping[f"{es_metric}"] = {}
                self.sk_attack_es[f"{es_metric}"] = {}
                for n in steps:
                    self.ge_search_early_stopping[f"{es_metric}"][f"{n}"] = []
                    self.sr_search_early_stopping[f"{es_metric}"][f"{n}"] = []
                    self.sk_attack_es[f"{es_metric}"][f"{n}"] = []
        else:
            if len(self.ge_search_early_stopping) == 0:
                self.ge_search_early_stopping[f"{es_metric}"] = []
                self.sr_search_early_stopping[f"{es_metric}"] = []
                self.sk_attack_es[f"{es_metric}"] = []

    def update_early_stopping_ge_and_sr(self, es_metric, ge, sr, n_profiling_traces):
        if self.settings["use_profiling_analyzer"]:
            self.ge_search_early_stopping[f"{es_metric}"][f"{n_profiling_traces}"].append(ge[-1])
            self.sr_search_early_stopping[f"{es_metric}"][f"{n_profiling_traces}"].append(sr[-1])
        else:
            self.ge_search_early_stopping[f"{es_metric}"].append(ge[-1])
            self.sr_search_early_stopping[f"{es_metric}"].append(sr[-1])

    def save_generic_metrics(self, profiling, history, search_index, hp_id, n_profiling_traces=None):
        if self.settings["use_early_stopping"]:
            early_stopping_metrics = self.get_early_stopping_metrics(
                profiling.get_builtin_callbacks()["early_stopping_callback"].get_metric_results())
            early_stopping_epoch_values = self.get_early_stopping_values_epochs(
                profiling.get_builtin_callbacks()["early_stopping_callback"].get_metric_results())
        else:
            early_stopping_metrics = None
            early_stopping_epoch_values = None
        self.analysis_db_controls.save_generic_metrics(history, early_stopping_metrics, early_stopping_epoch_values, hp_id,
                                                       search_index=search_index, n_profiling_traces=n_profiling_traces)

    def compute_metrics(self, model, search_index, rs_db_controls, builtin_callback, hp_id, n_profiling_traces=None, steps=None,
                        random_states=None):

        """ Compute SCA metrics: guessing entropy and success rate for attack and validation sets """

        """ 
        Compute Guessing Entropy, Success Rate and Number of Traces for Success (GE < 2) for attack and validation sets 
        Return scores/probabilities for each trace (p(l=leakage|trace=x))
        Retrieve random states for reproducibility
        Save results to database
        """

        label_name = f"Attack Set {search_index} {n_profiling_traces} traces" if self.settings[
            "use_profiling_analyzer"] else f"Attack Set {search_index}"
        random_states_ge_sr = random_states[f"{search_index}"][f"{label_name}"] if random_states is not None else None
        ge, sr, nt, sk, r = SCAMetrics(model, self.dataset.x_attack, self.settings, self.labels_key_guesses_attack_set).run(
            random_states=random_states_ge_sr)
        self.update_sk(sk, n_profiling_traces=n_profiling_traces)
        rs_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, search_index)

        label_name = f"Validation Set {search_index} {n_profiling_traces} traces" if self.settings[
            "use_profiling_analyzer"] else f"Validation Set {search_index}"
        random_states_ge_sr = random_states[f"{search_index}"][f"{label_name}"] if random_states is not None else None
        ge, sr, nt, sk, r = SCAMetrics(model, self.dataset.x_validation, self.settings, self.labels_key_guesses_validation_set).run(
            random_states=random_states_ge_sr)
        self.update_ge_and_sr(ge, sr, n_profiling_traces=n_profiling_traces)
        rs_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, search_index)

        """
        For Early Stopping Metrics:
        Compute Guessing Entropy, Success Rate and Number of Traces for Success (GE < 2) for attack and validation sets 
        Return scores/probabilities for each trace (p(l=leakage|trace=x))
        Retrieve random states for reproducibility        
        """

        """ TODO: search for saved h5 models in folder and then run metrics"""

        if self.settings["use_early_stopping"]:

            early_stopping_metrics = self.get_early_stopping_metrics(
                builtin_callback["early_stopping_callback"].get_metric_results())

            for es_metric in early_stopping_metrics:
                """ initialize list """
                self.initialize_early_stopping_ge_and_sr(es_metric, steps)

                self.load_model(model, es_metric)

                print(colored("Computing Guessing Entropy and Success Rate for Attack Set", "green"))
                label_name = f"ES Attack Set {es_metric} {search_index} {n_profiling_traces} traces" if self.settings[
                    "use_profiling_analyzer"] else f"ES Attack Set {es_metric} {search_index}"
                random_states_ge_sr = random_states[f"{search_index}"][f"{label_name}"] if random_states is not None else None
                ge, sr, nt, sk, r = SCAMetrics(model, self.dataset.x_attack, self.settings, self.labels_key_guesses_attack_set).run(
                    random_states=random_states_ge_sr)
                self.update_sk_es(sk, es_metric, n_profiling_traces)
                rs_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, search_index)

                print(colored("Computing Guessing Entropy and Success Rate for Validation Set", "green"))
                label_name = f"ES Validation Set {es_metric} {search_index} {n_profiling_traces} traces" if self.settings[
                    "use_profiling_analyzer"] else f"ES Validation Set {es_metric} {search_index}"
                random_states_ge_sr = random_states[f"{search_index}"][f"{label_name}"] if random_states is not None else None
                ge, sr, nt, sk, r = SCAMetrics(model, self.dataset.x_validation, self.settings, self.labels_key_guesses_validation_set).run(
                    random_states=random_states_ge_sr)
                rs_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, search_index)
                self.update_early_stopping_ge_and_sr(es_metric, ge, sr, n_profiling_traces)

                self.delete_model(es_metric)

    def finalize_search(self, profiling, model_descriptions, hp_ids, steps=None):

        """ Display results for best model """
        self.display_best_model()

        """ Update labels of best Guessing Entropy and Success Rate in database """
        if self.settings["use_early_stopping"]:
            early_stopping_metrics = self.get_early_stopping_metrics(
                profiling.get_builtin_callbacks()["early_stopping_callback"].get_metric_results())
        else:
            early_stopping_metrics = None
        if self.settings["use_profiling_analyzer"]:
            self.analysis_db_controls.update_best_model_metrics(self.best_model_index_pa, early_stopping_metrics, steps=steps)
        else:
            self.analysis_db_controls.update_best_model_metrics(self.best_model_index, early_stopping_metrics, steps=steps)

        """ Save model description (keras style) to database """
        if self.settings["use_profiling_analyzer"]:
            self.settings["best_hyperparameters_id"] = {}
            for n_profiling_traces in steps:
                best_model_index = self.best_model_index_pa[f"{n_profiling_traces}"]
                model_description = model_descriptions[best_model_index]
                self.analysis_db_controls.save_model_description_to_database(model_description, f"best_model_{best_model_index}",
                                                                             hp_ids[f"{n_profiling_traces}"][best_model_index])

                """ Update settings """
                self.settings["best_hyperparameters_id"][f"{n_profiling_traces}"] = hp_ids[f"{n_profiling_traces}"][
                    self.best_model_index_pa[f"{n_profiling_traces}"]]
            self.analysis_db_controls.db_update.db_update_settings(self.settings)
        else:
            best_model_index = self.best_model_index
            model_description = model_descriptions[best_model_index]
            self.analysis_db_controls.save_model_description_to_database(model_description, f"best_model_{best_model_index}",
                                                                         hp_ids[best_model_index])

            """ Update settings """
            self.settings["best_hyperparameters_id"] = hp_ids[int(best_model_index)]
            self.analysis_db_controls.db_update.db_update_settings(self.settings)

    def generate_model(self, hp_values):
        """ Generate a random/grid CNN or MLP model """
        if self.settings["use_grid_search"]:
            model, hp = self.generate_grid_model(hp_values)
        else:
            model, hp = self.generate_random_model(hp_values)
        self.hp_searches.append(hp)
        return model, hp



from aisy_sca.callbacks.CallbackControls import CallbackControls
from aisy_sca.metrics import SCAMetrics
import tensorflow.keras.backend as backend
from termcolor import colored
import numpy as np
import os


class Profiling:
    """
    Class to run profiling phase on the dataset and model
    model: keras model already compiled
    settings: analysis settings
    dataset: dataset structure containing traces, labels and metadata (plaintext, ciphertext, keys)
    search_index: for hyper-parameter search purposes
    train_best_model: for hyper-parameter search purposes
    """

    def __init__(self, settings, dataset, da_function):
        self.settings = settings
        self.dataset = dataset
        self.da_function = da_function
        self.builtin_callbacks = None
        self.custom_callbacks = None
        self.labels_key_guesses_attack_set = None
        self.labels_key_guesses_validation_set = None

    def train_model(self, model, n_profiling_traces=None):

        """
        Train the model (profiling phase).
        1. Configure dataset
        2. Train model
        """

        """ reshape traces if needed """
        input_layer_shape = model.get_layer(index=0).input_shape

        """ Check if model is created with Sequential or Model class from Keras """
        if len(input_layer_shape) == 1:
            """ Model was created with Model class """
            input_layer_shape = input_layer_shape[0]

        """ Check if neural network is a cnn or mlp """
        if len(input_layer_shape) == 3:
            self.dataset.reshape_for_cnn()
        else:
            self.dataset.reshape_for_mlp()

        """ Create callbacks """
        callback_controls = CallbackControls(self.dataset, self.settings)
        callback_controls.create_callbacks()
        callbacks = callback_controls.get_callbacks()
        self.builtin_callbacks = callback_controls.get_builtin_callbacks()
        self.custom_callbacks = callback_controls.get_custom_callbacks()

        if self.settings["split_test_set"] or self.settings["split_training_set"]:
            validation_set = (self.dataset.x_validation, self.dataset.y_validation)
        else:
            validation_set = (self.dataset.x_attack, self.dataset.y_attack)

        if self.settings["use_data_augmentation"]:

            """  If data augmentation is used, then x_profiling needs to be reshaped back to 2D. """

            x_profiling = self.dataset.x_profiling.reshape((self.dataset.x_profiling.shape[0], self.dataset.x_profiling.shape[1]))

            da_method = self.da_function(x_profiling, self.dataset.y_profiling, self.settings["batch_size"], input_layer_shape)
            history = model.fit_generator(
                generator=da_method,
                steps_per_epoch=self.settings["data_augmentation"][1],
                epochs=self.settings["epochs"],
                verbose=2,
                validation_data=validation_set,
                validation_steps=1,
                callbacks=callbacks)

        else:

            """ Train the model """

            history = model.fit(
                x=self.dataset.x_profiling[:n_profiling_traces] if self.settings["use_profiling_analyzer"] else self.dataset.x_profiling,
                y=self.dataset.y_profiling[:n_profiling_traces] if self.settings["use_profiling_analyzer"] else self.dataset.y_profiling,
                batch_size=self.settings["batch_size"],
                verbose=2,
                epochs=self.settings["epochs"],
                shuffle=True,
                validation_data=validation_set,
                callbacks=callbacks)

        return history

    def get_builtin_callbacks(self):
        return self.builtin_callbacks

    def get_custom_callbacks(self):
        return self.custom_callbacks

    def compute_metrics(self, model, analysis_db_controls, hp_id, n_profiling_traces=None, random_states=None):

        """ Compute SCA metrics: guessing entropy and success rate for attack and validation sets """

        """ 
        Compute Guessing Entropy, Success Rate and Number of Traces for Success (GE < 2) for attack and validation sets 
        Return scores/probabilities for each trace (p(l=leakage|trace=x))
        Retrieve random states for reproducibility
        Save results to database
        """

        label_name = f"Attack Set" if n_profiling_traces is None else f"Attack Set {n_profiling_traces} traces"
        random_states_ge_sr = random_states[f"{model['index']}"][f"{label_name}"] if random_states is not None else None
        ge, sr, nt, sk, r = SCAMetrics(model["model"], self.dataset.x_attack, self.settings, self.labels_key_guesses_attack_set).run(
            random_states=random_states_ge_sr)

        analysis_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, model['index'])

        if self.dataset.x_validation is not None:
            label_name = f"Validation Set" if n_profiling_traces is None else f"Validation Set {n_profiling_traces} traces"
            random_states_ge_sr = random_states[f"{model['index']}"][f"{label_name}"] if random_states is not None else None
            ge, sr, nt, sk, r = SCAMetrics(model["model"], self.dataset.x_validation, self.settings,
                                           self.labels_key_guesses_validation_set).run(
                random_states=random_states_ge_sr)
            analysis_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, model['index'])

        """
        For Early Stopping Metrics:
        Compute Guessing Entropy, Success Rate and Number of Traces for Success (GE < 2) for attack and validation sets 
        Return scores/probabilities for each trace (p(l=leakage|trace=x))
        Retrieve random states for reproducibility        
        """

        """ TODO: search for saved h5 models in folder and then run metrics"""

        if self.settings["use_early_stopping"]:

            early_stopping_metrics = self.get_early_stopping_metrics(
                self.builtin_callbacks["early_stopping_callback"].get_metric_results())

            for es_metric in early_stopping_metrics:
                self.load_model(model["model"], es_metric)

                print(colored("Computing Guessing Entropy and Success Rate for Attack Set", "green"))
                label_name = f"ES Attack Set {es_metric}" if n_profiling_traces is None else f"ES Attack Set {es_metric} {n_profiling_traces} traces"
                random_states_ge_sr = random_states[f"{model['index']}"][f"{label_name}"] if random_states is not None else None
                ge, sr, nt, sk, r = SCAMetrics(model["model"], self.dataset.x_attack, self.settings,
                                               self.labels_key_guesses_attack_set).run(
                    random_states=random_states_ge_sr)
                analysis_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, model['index'])

                print(colored("Computing Guessing Entropy and Success Rate for Validation Set", "green"))
                label_name = f"ES Validation Set {es_metric}" if n_profiling_traces is None else f"ES Validation Set {es_metric} {n_profiling_traces} traces"
                random_states_ge_sr = random_states[f"{model['index']}"][f"{label_name}"] if random_states is not None else None
                ge, sr, nt, sk, r = SCAMetrics(model["model"], self.dataset.x_validation, self.settings,
                                               self.labels_key_guesses_validation_set).run(
                    random_states=random_states_ge_sr)
                analysis_db_controls.save_sca_metrics_to_database(ge, sr, label_name, hp_id, r, model['index'])

                self.delete_model(es_metric)

    def load_model(self, model, setting, idx=None):

        """ Load weights from saved early stopping model from 'resource/models' directory"""

        if idx is None:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}.h5"
        else:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}_{idx}.h5"
        print(model_name)
        model.load_weights(model_name)

    def delete_model(self, setting, idx=None):

        """ Load model from 'resource/models' directory """

        if idx is None:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}.h5"
        else:
            model_name = f"{self.settings['resources_root_folder']}models/best_model_{setting}_{self.settings['timestamp']}_{idx}.h5"
        os.remove(model_name)

    def get_early_stopping_metrics(self, early_stopping_metric_results):
        early_stopping_metric_names = []
        for early_stopping_metric in self.settings["early_stopping"]["metrics"]:
            if isinstance(early_stopping_metric_results[early_stopping_metric][0], list):
                for idx in range(len(early_stopping_metric_results[early_stopping_metric][0])):
                    early_stopping_metric_names.append(f"{early_stopping_metric}_{idx}")
            else:
                early_stopping_metric_names.append(early_stopping_metric)
        return early_stopping_metric_names

    def get_early_stopping_values_epochs(self, early_stopping_metric_results):
        early_stopping_values_epochs = {}
        for early_stopping_metric in self.settings["early_stopping"]["metrics"]:
            if isinstance(early_stopping_metric_results[early_stopping_metric][0], list):
                for idx in range(len(early_stopping_metric_results[early_stopping_metric][0])):
                    early_stopping_values_epochs[f"{early_stopping_metric}_{idx}"] = np.array(
                        early_stopping_metric_results[early_stopping_metric])[:, idx]
            else:
                early_stopping_values_epochs[f"{early_stopping_metric}"] = early_stopping_metric_results[early_stopping_metric]
        return early_stopping_values_epochs



from aisy_sca.optimization.HyperparameterSearch import *
from aisy_sca.sca.Profiling import Profiling
from aisy_sca.utils import Utils
import tensorflow.keras.backend as backend
from termcolor import colored
import itertools
import json
import time


class RandomGridSearch(HyperparameterSearch):
    """
    Class that perform random or grid hyper-parameter search .
    The random search only supports MLPs and CNNs.
    """

    def __init__(self, settings, dataset, random_search_definitions, labels_key_guesses_attack_set, labels_key_guesses_validation_set,
                 analysis_db_controls, random_states=None):
        super().__init__(settings, dataset, random_search_definitions, labels_key_guesses_attack_set, labels_key_guesses_validation_set,
                         analysis_db_controls, random_states=None)
        self.custom_callbacks = {}

    def run(self, hp_combinations_reproducible, da_function=None, grid_search=False):

        hp_search_ranges = self.random_search_definitions["hyper_parameters_search"]
        hp_ids = []
        model_descriptions = []
        if self.random_states is None:
            if grid_search:
                hp_to_search = {}
                hp_to_adjust = {}
                for hp, hp_value in hp_search_ranges.items():
                    if type(hp_value) is str:
                        hp_to_adjust[hp] = hp_value
                    else:
                        hp_to_search[hp] = hp_value
                keys, value = zip(*hp_to_search.items())
                search_hp_combinations = [dict(zip(keys, v)) for v in itertools.product(*value)]
                for idx in range(len(search_hp_combinations)):
                    for hp, hp_value in hp_to_adjust.items():
                        search_hp_combinations[idx][hp] = hp_value

                self.random_search_definitions["max_trials"] = len(search_hp_combinations)
                print("======================================================================================")
                print(f"Number of hyperparameters combinations for grid search: {len(search_hp_combinations)}")
                print("======================================================================================")
            else:
                search_hp_combinations = hp_search_ranges
        else:
            search_hp_combinations = hp_combinations_reproducible

        start = time.time()

        for search_index in range(self.random_search_definitions["max_trials"]):

            """ Generate a random/grid CNN or MLP model """
            if self.random_states is None:
                hp_values = search_hp_combinations[search_index] if grid_search else search_hp_combinations
            else:
                hp_values = search_hp_combinations[search_index]
            model, hp = self.generate_model(hp_values)

            print(colored(f"Hyper-Parameters for Search {search_index}: {json.dumps(hp, sort_keys=True, indent=4)}", "blue"))

            model_descriptions.append(Utils().keras_model_as_string(model, f"best_model_{search_index}"))

            """ Run profiling/training phase """
            profiling = Profiling(self.settings, self.dataset, da_function)
            history = profiling.train_model(model)
            self.custom_callbacks[f"{search_index}"] = profiling.get_custom_callbacks()

            """ Save hyperparameters combination to database and receive the id of the insertion (hp_id)"""
            hyper_parameters_list = self.set_hyper_parameters_list(hp, model)
            hp_id = self.analysis_db_controls.save_hyper_parameters_to_database(hyper_parameters_list)
            hp_ids.append(hp_id)

            """ Retrieve metrics from model and save them to database """
            self.loss_search.append(history.history["loss"])
            self.acc_search.append(history.history["accuracy"])

            """ Compute SCA metrics and save them to database """
            self.compute_metrics(model, search_index, self.analysis_db_controls, profiling.get_builtin_callbacks(), hp_id)
            ge = self.ge_search[search_index]
            sr = self.sr_search[search_index]
            hyper_parameters_list = self.update_hyper_parameters_list(hp, ge, sr)
            self.analysis_db_controls.update_hyper_parameters_to_database(hyper_parameters_list, hp_id)

            """ Save generic metrics (loss, accuracy, early stopping metrics) to database """
            self.save_generic_metrics(profiling, history, search_index, hp_id)

            """ Update results in database """
            self.analysis_db_controls.update_results_in_database(time.time() - start)

            backend.clear_session()

            """ Find best model"""
            self.get_best_model()

            """ Save best model so far in h5 """
            self.save_best_model(model, search_index)

            """ Check if stopping condition is satisfied"""
            if self.check_stop_condition(search_index, hp):
                self.finalize_search(profiling, model_descriptions, hp_ids)
                break

            if search_index == self.random_search_definitions["max_trials"] - 1:
                self.finalize_search(profiling, model_descriptions, hp_ids)

        """ update database settings"""
        self.analysis_db_controls.update_results_in_database(time.time() - start)

    def get_custom_callbacks(self):
        return self.custom_callbacks



from aisy_sca.optimization.RandomGridSearch import *
from aisy_sca.optimization.HyperparameterSearch import *
from aisy_sca.optimization.KerasCNNRandom import KerasCNNRandom
from aisy_sca.optimization.KerasMLPRandom import KerasMLPRandom
from aisy_sca.optimization.KerasCNNGrid import KerasCNNGrid
from aisy_sca.optimization.KerasMLPGrid import KerasMLPGrid
from aisy_sca.optimization.Ensemble import Ensemble
from aisy_sca.optimization.EnsembleControls import EnsembleControls
from aisy_sca.optimization.ProfilingAnalyzer import *



from aisy_sca.analysis.SingleProcess import SingleProcess
from aisy_sca.analysis.HyperparameterSearchProcess import HyperparameterSearchProcess
from aisy_sca.analysis.SingleModel import SingleModel
from aisy_sca.analysis.MultiModel import MultiModel



import numpy as np
import scipy.special as scipy_special


class SCAMetrics:

    def __init__(self, model, x_data, settings, label_key_guess):
        self.model = model
        self.x_data = x_data
        self.settings = settings
        self.label_key_guess = label_key_guess

    def run(self, random_states=None):

        nt = len(self.x_data)
        nt_interval = int(self.settings["key_rank_attack_traces"] / self.settings["key_rank_report_interval"])
        key_ranking_sum = np.zeros(nt_interval)
        success_rate_sum = np.zeros(nt_interval)
        r_states = []

        # ---------------------------------------------------------------------------------------------------------#
        # predict output probabilities for shuffled test or validation set
        # ---------------------------------------------------------------------------------------------------------#
        output_probabilities = self.model.predict(self.x_data)
        if self.model.layers[len(self.model.layers) - 1].get_config()['activation'] is not "softmax":
            output_probabilities = scipy_special.softmax(output_probabilities, axis=1)

        nb_guesses = len(self.label_key_guess)
        probabilities_kg_all_traces = np.zeros((nt, nb_guesses))
        for index in range(nt):
            probabilities_kg_all_traces[index] = output_probabilities[index][
                np.asarray([int(leakage[index]) for leakage in self.label_key_guess[:]])  # array with 256 leakage values (1 per key guess)
            ]

        for run in range(self.settings["key_rank_executions"]):

            if random_states is None:
                r_state = np.random.randint(1048576)
                np.random.seed(r_state)
                r = np.random.choice(nt, self.settings["key_rank_attack_traces"], replace=False)
            else:
                r_state = random_states[f"{run}"]
                np.random.seed(r_state)
                r = np.random.choice(nt, self.settings["key_rank_attack_traces"], replace=False)
            r_states.append(r_state)

            probabilities_kg_all_traces_shuffled = probabilities_kg_all_traces[r]

            key_probabilities = np.zeros(nb_guesses)

            kr_count = 0
            for index in range(self.settings["key_rank_attack_traces"]):

                key_probabilities += np.log(probabilities_kg_all_traces_shuffled[index] + 1e-36)
                key_probabilities_sorted = np.argsort(key_probabilities)[::-1]

                if (index + 1) % self.settings["key_rank_report_interval"] == 0:
                    key_ranking_good_key = list(key_probabilities_sorted).index(self.settings["good_key"]) + 1
                    key_ranking_sum[kr_count] += key_ranking_good_key

                    if key_ranking_good_key == 1:
                        success_rate_sum[kr_count] += 1

                    kr_count += 1

        guessing_entropy = key_ranking_sum / self.settings["key_rank_executions"]
        success_rate = success_rate_sum / self.settings["key_rank_executions"]
        if guessing_entropy[nt_interval - 1] < 2:
            result_number_of_traces_ge_1 = self.settings["key_rank_attack_traces"] - np.argmax(guessing_entropy[::-1] > 2) * self.settings[
                "key_rank_report_interval"]
        else:
            result_number_of_traces_ge_1 = self.settings["key_rank_executions"]

        print(f"GE = {guessing_entropy[nt_interval - 1]}")
        print(f"SR = {success_rate[nt_interval - 1]}")
        if guessing_entropy[nt_interval - 1] < 2:
            print(f"Number of traces to reach GE = 1: {result_number_of_traces_ge_1}")

        return guessing_entropy, success_rate, result_number_of_traces_ge_1, output_probabilities, r_states



from aisy_sca.sca.Profiling import Profiling
from aisy_sca.utils.utils import Utils
import tensorflow.keras.backend as backend
import time


class SingleModel(Profiling):

    def __init__(self, settings, dataset, da_function, random_states=None):
        super().__init__(settings, dataset, da_function)
        self.settings = settings
        self.dataset = dataset
        self.da_function = da_function
        self.labels_key_guesses_attack_set = None
        self.labels_key_guesses_validation_set = None
        self.analysis_db_controls = None
        self.custom_callbacks = None
        self.random_states = random_states

    def run(self, model):

        initial_weights = model["model"].get_weights()
        model["model"].summary()

        start = time.time()

        hp_id = None

        for idx, n_profiling in enumerate(self.settings["profiling_analyzer_steps"]):

            """ Run profiling/training phase """

            model["model"].set_weights(initial_weights)
            history = self.train_model(model["model"], n_profiling_traces=n_profiling)
            self.custom_callbacks = self.get_custom_callbacks()

            """ Save hyper-parameters combination to database """
            if idx == 0:
                hyper_parameters_list = Utils().get_hyperparameters_from_model(model["model"])
                hp_id = self.analysis_db_controls.save_hyper_parameters_to_database(hyper_parameters_list)

            """ Retrieve metrics from model and save them to database """
            if self.settings["use_early_stopping"]:
                early_stopping_metrics = self.get_early_stopping_metrics(
                    self.get_builtin_callbacks()["early_stopping_callback"].get_metric_results())
                early_stopping_epoch_values = self.get_early_stopping_values_epochs(
                    self.get_builtin_callbacks()["early_stopping_callback"].get_metric_results())
            else:
                early_stopping_metrics = None
                early_stopping_epoch_values = None
            self.analysis_db_controls.save_generic_metrics(history, early_stopping_metrics, early_stopping_epoch_values, hp_id,
                                                           n_profiling_traces=n_profiling if self.settings[
                                                               "use_profiling_analyzer"] else None)

            """ Compute SCA metrics and save them to database """
            self.compute_metrics(model, self.analysis_db_controls, hp_id,
                                 n_profiling_traces=n_profiling if self.settings["use_profiling_analyzer"] else None,
                                 random_states=self.random_states)

            """ Save model description (keras style) to database """
            model_description = Utils().keras_model_as_string(model["model"], model["method_name"])
            self.analysis_db_controls.save_model_description_to_database(model_description, model["method_name"], hp_id)

            """ Save visualization results to database """
            if self.settings["use_visualization"]:
                self.analysis_db_controls.save_visualization_results_to_database(
                    self.get_builtin_callbacks()["input_gradients_callback"].input_gradients_epochs(),
                    self.get_builtin_callbacks()["input_gradients_callback"].input_gradients_sum(), hp_id)

            """ Save confusion matrix results to database """
            if self.settings["use_confusion_matrix"]:
                self.analysis_db_controls.save_confusion_matrix_results_to_database(
                    self.get_builtin_callbacks()["confusion_matrix_callback"].get_confusion_matrix(), hp_id)

            """ update database settings"""
            self.analysis_db_controls.update_results_in_database(time.time() - start)

            backend.clear_session()



from aisy_sca.crypto.aes.aes128 import *


def aes_intermediates_decrypt_from_input(ciphertexts, keys, leakage_model):
    if leakage_model["leakage_model"] in ["HD", "HD_ID", "HD_bit"]:
        return get_hd_intermediates_decrypt_input(ciphertexts, keys, leakage_model)
    elif leakage_model["leakage_model"] in ["HW", "ID", "bit"]:
        return get_hw_id_intermediates_decrypt_input(ciphertexts, keys, leakage_model)
    else:
        print("Leakage model not supported!")
        return None


def aes_intermediates_decrypt_from_output(plaintexts, keys, leakage_model):
    if leakage_model["leakage_model"] in ["HD", "HD_ID", "HD_bit"]:
        return get_hd_intermediates_decrypt_output(plaintexts, keys, leakage_model)
    elif leakage_model["leakage_model"] in ["HW", "ID", "bit"]:
        return get_hw_id_intermediates_decrypt_output(plaintexts, keys, leakage_model)
    else:
        print("Leakage model not supported!")
        return None


def get_input_xor_sbox_out_decrypt_input(leakage_model, ciphertexts, keys):
    state1 = [row[inv_shift_row_mask[leakage_model["byte"]]] for row in ciphertexts]
    ciphertext = [row[leakage_model["byte"]] for row in ciphertexts]
    key = [row[10][leakage_model["byte"]] for row in keys]
    return [inv_s_box[int(c1) ^ int(k)] ^ int(c2) for c1, c2, k in zip(np.asarray(state1[:]), np.asarray(ciphertext[:]),
                                                                       np.asarray(key[:]))]


def get_sbox_in_xor_sbox_out_decrypt_input(leakage_model, ciphertexts, keys):
    ciphertext = [row[inv_shift_row_mask[leakage_model["byte"]]] for row in ciphertexts]
    key = [row[10][leakage_model["byte"]] for row in keys]
    state1 = [int(c) ^ int(k) for c, k in zip(np.asarray(ciphertext[:]), np.asarray(key[:]))]
    state2 = inv_s_box[state1]
    return [int(s1) ^ int(s2) for s1, s2 in zip(np.asarray(state1[:]), np.asarray(state2[:]))]


def get_addroundkey_out_decrypt_input(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_add_round_key = []
    for ciphertext, key in zip(ciphertexts, keys):
        state_add_round_key.append(add_round_key(ciphertext, key[10]))
    return [int(s[leakage_model["byte"]]) for s in state_add_round_key]


def get_inv_shiftrows_out_decrypt_input(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_add_round_key = []
    for ciphertext, key in zip(ciphertexts, keys):
        state_add_round_key.append(add_round_key(ciphertext, key[10]))
    return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_add_round_key]


def get_inv_sbox_out_decrypt_input(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_inv_sbox = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state_inv_sbox.append(inv_sub_bytes(state))
    return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_inv_sbox]


def get_addroundkey_out_decrypt_input_round_2(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_add_round_key = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state = inv_shift_rows(state)
        state = inv_sub_bytes(state)
        state_add_round_key.append(add_round_key(state, key[9]))  # todo: compute the constant value for rk9
        return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_add_round_key]


def get_inv_mixcolumns_out_decrypt_input_round_2(leakage_model, ciphertexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    ciphertexts = np.array(ciphertexts, dtype=np.uint8)
    state_mc = []
    for ciphertext, key in zip(ciphertexts, keys):
        state = add_round_key(ciphertext, key[10])
        state = inv_shift_rows(state)
        state = inv_sub_bytes(state)
        state = add_round_key(state, key[9])  # todo: compute the constant value for rk9
        state_mc.append(inv_mix_columns(state))
    return [int(s[shift_row_mask[leakage_model["byte"]]]) for s in state_mc]


def get_hd_intermediates_decrypt_input(ciphertexts, keys, leakage_model):
    """
    There is support for the following leakage models:
    - HD(SboxIn XOR Ciphertext) or SboxIn XOR Ciphertext
    - HD(SboxIn XOR SboxOut) or SboxIn XOR SboxOut
    """
    if leakage_model["round_first"] == 1:
        if leakage_model["target_state_first"] == "Input" and leakage_model["target_state_second"] == "InvSbox":
            return get_input_xor_sbox_out_decrypt_input(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state_first"] == "AddRoundKey" and leakage_model["target_state_second"] == "InvSbox":
            return get_sbox_in_xor_sbox_out_decrypt_input(leakage_model, ciphertexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None


def get_hw_id_intermediates_decrypt_input(ciphertexts, keys, leakage_model):
    if leakage_model["round"] == 1:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_out_decrypt_input(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "InvShiftRows":
            return get_inv_shiftrows_out_decrypt_input(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "InvSbox":
            return get_inv_sbox_out_decrypt_input(leakage_model, ciphertexts, keys)
    elif leakage_model["round"] == 2:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_out_decrypt_input_round_2(leakage_model, ciphertexts, keys)
        elif leakage_model["target_state"] is "InvMixColumns":
            return get_inv_mixcolumns_out_decrypt_input_round_2(leakage_model, ciphertexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None


def get_output_xor_sbox_in_decrypt_output(leakage_model, plaintexts, keys):
    state1 = [row[leakage_model["byte"]] for row in plaintexts]
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [s_box[int(p1) ^ int(k)] ^ int(p2) for p1, p2, k in zip(np.asarray(state1[:]), np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_sbox_in_xor_sbox_out_decrypt_output(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    state1 = [int(p) ^ int(k) for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]
    state2 = s_box[state1]
    return [int(s1) ^ int(s2) for s1, s2 in zip(np.asarray(state1[:]), np.asarray(state2[:]))]


def get_addroundkey_in_decrypt_output(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [int(p) ^ int(k) for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_inv_sbox_in_decrypt_output(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [s_box[int(p) ^ int(k)] for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_inv_shiftrows_in_decrypt_output(leakage_model, plaintexts, keys):
    plaintext = [row[leakage_model["byte"]] for row in plaintexts]
    key = [row[leakage_model["byte"]] for row in keys]
    return [inv_s_box[int(p) ^ int(k)] for p, k in zip(np.asarray(plaintext[:]), np.asarray(key[:]))]


def get_inv_mixcolumns_in_decrypt_output_round_9(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_mc = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, round_keys[0])
        state = sub_bytes(state)
        state = shift_rows(state)
        state_mc.append(mix_columns(state))
        return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_mc]


def get_addroundkey_in_decrypt_output_round_9(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_add_round_key = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, round_keys[0])
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state_add_round_key.append(add_round_key(state, round_keys[1]))  # todo: compute the constant value for rk1
        return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_add_round_key]


def get_inv_sbox_in_decrypt_output_round_9(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_sbox = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, round_keys[0])
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state = add_round_key(state, round_keys[1])  # todo: compute the constant value for rk1
        state_sbox.append(sub_bytes(state))
        return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_sbox]


def get_inv_shiftrows_in_decrypt_output_round_9(leakage_model, plaintexts, keys):
    keys = np.array(keys, dtype=np.uint8)
    plaintexts = np.array(plaintexts, dtype=np.uint8)
    state_sbox = []
    for plaintext, key in zip(plaintexts, keys):
        round_keys = expand_key(key.copy())
        state = add_round_key(plaintext, round_keys[0])
        state = sub_bytes(state)
        state = shift_rows(state)
        state = mix_columns(state)
        state = add_round_key(state, round_keys[1])  # todo: compute the constant value for rk1
        state_sbox.append(sub_bytes(state))
        return [int(s[inv_shift_row_mask[leakage_model["byte"]]]) for s in state_sbox]


def get_hd_intermediates_decrypt_output(plaintexts, keys, leakage_model):
    """
    There is support for the following leakage models:
    - HD(SboxIn XOR Ciphertext) or SboxIn XOR Ciphertext
    - HD(SboxIn XOR SboxOut) or SboxIn XOR SboxOut
    """
    if leakage_model["round_first"] == 1:
        if leakage_model["target_state_first"] == "Output" and leakage_model["target_state_second"] == "InvSbox":
            return get_output_xor_sbox_in_decrypt_output(leakage_model, plaintexts, keys)
        elif leakage_model["target_state_first"] == "AddRoundKey" and leakage_model["target_state_second"] == "InvSbox":
            return get_sbox_in_xor_sbox_out_decrypt_output(leakage_model, plaintexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None


def get_hw_id_intermediates_decrypt_output(plaintexts, keys, leakage_model):
    if leakage_model["round"] == 10:
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_in_decrypt_output(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "InvSbox":
            return get_inv_sbox_in_decrypt_output(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "InvShiftRows":
            return get_inv_shiftrows_in_decrypt_output(leakage_model, plaintexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    elif leakage_model["round"] == 9:
        if leakage_model["target_state"] is "InvMixColumns":
            return get_inv_mixcolumns_in_decrypt_output_round_9(leakage_model, plaintexts, keys)
        if leakage_model["target_state"] is "AddRoundKey":
            return get_addroundkey_in_decrypt_output_round_9(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "InvSbox":
            return get_inv_sbox_in_decrypt_output_round_9(leakage_model, plaintexts, keys)
        elif leakage_model["target_state"] is "InvShiftRows":
            return get_inv_shiftrows_in_decrypt_output_round_9(leakage_model, plaintexts, keys)
        else:
            print("Leakage model not supported!")
            return None
    else:
        print("Leakage model not supported!")
        return None



import numpy as np


class EnsembleControls:

    def __init__(self, settings):
        self.settings = settings

    def get_nt_for_ge_1(self, ge):
        return self.settings["key_rank_attack_traces"] - np.argmax(ge[::-1] > 2) * self.settings["key_rank_report_interval"]

    def get_nt_for_sr_thre(self, sr, thre):
        return self.settings["key_rank_attack_traces"] - np.argmax(sr[::-1] < thre) * self.settings["key_rank_report_interval"]

    def order_models_per_guessing_entropy(self, ge_values, number_of_models):

        """ Order models by the number of validation traces required to achieve GE = 1"""

        result_number_of_traces_ge_1 = []
        for model_index, ge in enumerate(ge_values):
            nt_ge_1 = self.get_nt_for_ge_1(ge) if ge[len(ge) - 1] < 2 else self.settings["key_rank_attack_traces"]
            result_number_of_traces_ge_1.append([ge[len(ge) - 1], nt_ge_1, model_index])

        sorted_models = sorted(result_number_of_traces_ge_1, key=lambda l: l[:])

        list_of_best_models = []
        for model_index in range(number_of_models):
            list_of_best_models.append(sorted_models[model_index][2])

        return list_of_best_models

    def order_models_per_success_rate(self, sr_values, thre, number_of_best_models):

        result_number_of_traces_sr_1 = []
        for model_index, sr in enumerate(sr_values):
            nt_sr_1 = self.get_nt_for_sr_thre(sr, thre) if sr[len(sr) - 1] > thre else self.settings["key_rank_attack_traces"]
            result_number_of_traces_sr_1.append([sr[len(sr) - 1], nt_sr_1, model_index])

        sorted_models = sorted(result_number_of_traces_sr_1, key=lambda l: l[:])

        list_of_best_models = []
        for model_index in range(number_of_best_models):
            list_of_best_models.append(sorted_models[model_index][2])

        return list_of_best_models



from aisy_database.db_update import DBUpdate
from aisy_database.db_save import DBSave
from aisy_database.db_select import DBSelect
from aisy_database.db_tables import *


class AnalysisDatabaseControls:

    def __init__(self, settings):
        self.settings = settings
        self.db_save = DBSave(f"{settings['database_root_folder']}{settings['database_name']}")
        self.analysis_id = self.db_save.db_save_analysis(settings['database_name'], settings["filename"], settings)
        self.db_save.analysis_id = self.analysis_id
        self.db_update = DBUpdate(f"{settings['database_root_folder']}{settings['database_name']}", self.analysis_id)
        self.db_select = DBSelect(f"{settings['database_root_folder']}{settings['database_name']}")

    def update_results_in_database(self, elapsed_time):
        self.db_update.db_update_settings(self.settings)
        self.db_update.db_update_elapsed_time(elapsed_time)

    def save_leakage_model_to_database(self):
        if "leakage_model" in self.settings:
            leakage_model = {
                "cipher": self.settings["leakage_model"]["cipher"],
                "leakage_model": self.settings["leakage_model"]["leakage_model"],
                "byte": self.settings["leakage_model"]["byte"],
                "round": self.settings["leakage_model"]["round"],
                "target_state": self.settings["leakage_model"]["target_state"],
                "direction": self.settings["leakage_model"]["direction"],
                "attack_direction": self.settings["leakage_model"]["attack_direction"]
            }
            if self.settings["leakage_model"]["leakage_model"] == "bit":
                leakage_model["bit"] = self.settings["leakage_model"]["bit"]

            if self.settings["leakage_model"]["leakage_model"] == "HD":
                leakage_model["round_first"] = self.settings["leakage_model"]["round_first"]
                leakage_model["round_second"] = self.settings["leakage_model"]["round_second"]
                leakage_model["target_state_first"] = self.settings["leakage_model"]["target_state_first"]
                leakage_model["target_state_second"] = self.settings["leakage_model"]["target_state_second"]
        else:
            leakage_model = {"leakage_model": "Unknown", "cipher": "Unknown", "byte": "Unknown"}

        self.db_save.db_save_leakage_model(leakage_model)

    def save_visualization_results_to_database(self, input_gradients_epochs, input_gradients_sum, hp_id):
        self.db_save.db_save_visualization(input_gradients_epochs, input_gradients_sum, f"Input Gradients {hp_id}", hp_id)

    def save_confusion_matrix_results_to_database(self, confusion_matrix, hp_id):
        self.db_save.db_save_confusion_matrix(confusion_matrix, hp_id)

    def save_sca_metrics_to_database(self, ge, sr, db_metric_name, hp_id, r, index):
        metric_id = self.db_save.db_save_key_rank(ge, db_metric_name, self.settings["key_rank_report_interval"])
        self.db_save.db_save_guessing_entropy_hyperparameter(metric_id, hp_id)
        metric_id = self.db_save.db_save_success_rate(sr, db_metric_name, self.settings["key_rank_report_interval"])
        self.db_save.db_save_success_rate_hyperparameter(metric_id, hp_id)
        self.db_save.db_save_random_state_hyperparameter(db_metric_name, index, r)

    def save_model_description_to_database(self, model_description, model_name, hyperparameters_id):
        self.db_save.db_save_neural_network(model_description, model_name, hyperparameters_id)

    def save_hyper_parameters_to_database(self, hyper_parameters_list):
        return self.db_save.db_save_hyper_parameters(hyper_parameters_list)

    def update_hyper_parameters_to_database(self, hyper_parameters_list, hp_id):
        self.db_update.db_update_hyperparameters(hyper_parameters_list, hp_id, "hyperparameters")

    def update_guessing_entropy_label(self, old_label, new_label):
        self.db_update.db_update_key_rank_label(old_label, new_label)

    def update_success_rate_label(self, old_label, new_label):
        self.db_update.db_update_success_rate_label(old_label, new_label)

    def update_metric_label(self, old_label, new_label):
        self.db_update.db_update_metric_label(old_label, new_label)

    def update_random_state_label(self, old_label, new_label):
        self.db_update.db_update_random_state_label(old_label, new_label)

    def save_generic_metrics(self, history, early_stopping_metrics, early_stopping_epoch_values, hp_id, search_index=None,
                             n_profiling_traces=None):
        if self.settings["use_profiling_analyzer"]:
            metric_id = self.db_save.db_save_metric(history.history["loss"],
                                                    f"loss {n_profiling_traces} traces" if search_index is None else f"loss {search_index} {n_profiling_traces} traces")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["accuracy"],
                                                    f"accuracy {n_profiling_traces} traces" if search_index is None else f"accuracy {search_index} {n_profiling_traces} traces")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["val_loss"],
                                                    f"val_loss {n_profiling_traces} traces" if search_index is None else f"val_loss {search_index} {n_profiling_traces} traces")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["val_accuracy"],
                                                    f"val_accuracy {n_profiling_traces} traces" if search_index is None else f"val_accuracy {search_index} {n_profiling_traces} traces")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            if early_stopping_metrics is not None and early_stopping_epoch_values is not None:
                for es_metric in early_stopping_metrics:
                    metric_id = self.db_save.db_save_metric(early_stopping_epoch_values[es_metric],
                                                            f"val_{es_metric} {n_profiling_traces} traces" if search_index is None else f"val_{es_metric} {search_index} {n_profiling_traces} traces")
                    self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
        else:
            metric_id = self.db_save.db_save_metric(history.history["loss"], "loss" if search_index is None else f"loss {search_index}")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["accuracy"],
                                                    "accuracy" if search_index is None else f"accuracy {search_index}")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["val_loss"],
                                                    "val_loss" if search_index is None else f"val_loss {search_index}")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            metric_id = self.db_save.db_save_metric(history.history["val_accuracy"],
                                                    "val_accuracy" if search_index is None else f"val_accuracy {search_index}")
            self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)
            if early_stopping_metrics is not None and early_stopping_epoch_values is not None:
                for es_metric in early_stopping_metrics:
                    metric_id = self.db_save.db_save_metric(early_stopping_epoch_values[es_metric],
                                                            f"val_{es_metric}" if search_index is None else f"val_{es_metric} {search_index}")
                    self.db_save.db_save_metric_hyperparameter(metric_id, hp_id)

    def update_best_model_metrics(self, best_model_index, early_stopping_metrics, steps=None):

        """ Update label of best Guessing Entropy and Success Rate in database """

        if self.settings["use_profiling_analyzer"]:
            for n in steps:
                self.update_guessing_entropy_label(f"Attack Set {best_model_index[f'{n}']} {n} traces",
                                                   f"Attack Set Best Model {n} traces")
                self.update_guessing_entropy_label(f"Validation Set {best_model_index[f'{n}']} {n} traces",
                                                   f"Validation Set Best Model {n} traces")
                self.update_success_rate_label(f"Attack Set {best_model_index[f'{n}']} {n} traces",
                                               f"Attack Set Best Model {n} traces")
                self.update_success_rate_label(f"Validation Set {best_model_index[f'{n}']} {n} traces",
                                               f"Validation Set Best Model {n} traces")
                self.update_metric_label(f"accuracy {best_model_index[f'{n}']} {n} traces",
                                         f"accuracy Best Model {n} traces")
                self.update_metric_label(f"loss {best_model_index[f'{n}']} {n} traces",
                                         f"loss Best Model {n} traces")
                self.update_metric_label(f"val_accuracy {best_model_index[f'{n}']} {n} traces",
                                         f"val_accuracy Best Model {n} traces")
                self.update_metric_label(f"val_loss {best_model_index[f'{n}']} {n} traces",
                                         f"val_loss Best Model {n} traces")
                if early_stopping_metrics is not None:
                    for es_metric in early_stopping_metrics:
                        self.update_guessing_entropy_label(f"ES Attack Set {es_metric} {best_model_index[f'{n}']} {n} traces",
                                                           f"ES Attack Set {es_metric} Best Model {n} traces")
                        self.update_guessing_entropy_label(f"ES Validation Set {es_metric} {best_model_index[f'{n}']} {n} traces",
                                                           f"ES Validation Set {es_metric} Best Model {n} traces")
                        self.update_success_rate_label(f"ES Attack Set {es_metric} {best_model_index[f'{n}']} {n} traces",
                                                       f"ES Attack Set {es_metric} Best Model {n} traces")
                        self.update_success_rate_label(f"ES Validation Set {es_metric} {best_model_index[f'{n}']} {n} traces",
                                                       f"ES Validation Set {es_metric} Best Model {n} traces")
                        self.update_metric_label(f"val_{es_metric} {best_model_index[f'{n}']} {n} traces",
                                                 f"val_{es_metric} Best Model {n} traces")
        else:
            self.update_guessing_entropy_label(f"Attack Set {best_model_index}", "Attack Set Best Model")
            self.update_guessing_entropy_label(f"Validation Set {best_model_index}", "Validation Set Best Model")
            self.update_success_rate_label(f"Attack Set {best_model_index}", "Attack Set Best Model")
            self.update_success_rate_label(f"Validation Set {best_model_index}", "Validation Set Best Model")
            self.update_metric_label(f"accuracy {best_model_index}", "accuracy Best Model")
            self.update_metric_label(f"loss {best_model_index}", "loss Best Model")
            self.update_metric_label(f"val_accuracy {best_model_index}", "val_accuracy Best Model")
            self.update_metric_label(f"val_loss {best_model_index}", "val_loss Best Model")
            if early_stopping_metrics is not None:
                for es_metric in early_stopping_metrics:
                    self.update_guessing_entropy_label(f"ES Attack Set {es_metric} {best_model_index}",
                                                       f"ES Attack Set {es_metric} Best Model")
                    self.update_guessing_entropy_label(f"ES Validation Set {es_metric} {best_model_index}",
                                                       f"ES Validation Set {es_metric} Best Model")
                    self.update_success_rate_label(f"ES Attack Set {es_metric} {best_model_index}",
                                                   f"ES Attack Set {es_metric} Best Model")
                    self.update_success_rate_label(f"ES Validation Set {es_metric} {best_model_index}",
                                                   f"ES Validation Set {es_metric} Best Model")
                    self.update_metric_label(f"val_{es_metric} {best_model_index}", f"val_{es_metric} Best Model")

    def select_all_guessing_entropy_from_analysis(self, n_profiling=None):
        ge_values = self.db_select.select_all_guessing_entropy_from_analysis(GuessingEntropy, self.settings["analysis_id"])

        ge_validation = []
        ge_attack = []

        for ge in ge_values:
            self.append_result(ge_validation, ge, "Validation", n_profiling)
            self.append_result(ge_attack, ge, "Attack", n_profiling)

        return ge_validation, ge_attack

    def select_all_success_rate_from_analysis(self, n_profiling=None):
        sr_values = self.db_select.select_all_success_rate_from_analysis(SuccessRate, self.settings["analysis_id"])

        sr_validation = []
        sr_attack = []
        for sr in sr_values:
            self.append_result(sr_validation, sr, "Validation", n_profiling)
            self.append_result(sr_attack, sr, "Attack", n_profiling)

        return sr_validation, sr_attack

    def select_all_early_stopping_guessing_entropy_from_analysis(self, early_stopping_metric, n_profiling=None):
        ge_values = self.db_select.select_all_guessing_entropy_from_analysis(GuessingEntropy, self.settings["analysis_id"])

        ge_validation = []
        ge_attack = []
        for ge in ge_values:
            self.append_result_early_stopping(ge_validation, ge, "Validation", n_profiling, early_stopping_metric)
            self.append_result_early_stopping(ge_attack, ge, "Attack", n_profiling, early_stopping_metric)

        return ge_validation, ge_attack

    def select_all_early_stopping_success_rate_from_analysis(self, early_stopping_metric, n_profiling=None):
        sr_values = self.db_select.select_all_success_rate_from_analysis(SuccessRate, self.settings["analysis_id"])

        sr_validation = []
        sr_attack = []
        for sr in sr_values:
            self.append_result_early_stopping(sr_validation, sr, "Validation", n_profiling, early_stopping_metric)
            self.append_result_early_stopping(sr_attack, sr, "Attack", n_profiling, early_stopping_metric)

        return sr_validation, sr_attack

    def append_result(self, result_list, result, label_set, n_profiling):
        if label_set in result["label"] and "ES" not in result["label"]:
            if n_profiling is not None:
                if f"{n_profiling} traces" in result["label"]:
                    result_list.append(result["values"])
            else:
                result_list.append(result["values"])

    def append_result_early_stopping(self, result_list, result, label_set, n_profiling, early_stopping_metric):
        if label_set in result["label"] and "ES" in result["label"] and early_stopping_metric in result["label"]:
            if n_profiling is not None:
                if f"{n_profiling} traces" in result["label"]:
                    result_list.append(result["values"])
            else:
                result_list.append(result["values"])

    def save_ensemble_guessing_entropy_results(self, ge, label):
        self.db_save.db_save_key_rank(ge, label, self.settings["key_rank_report_interval"])

    def save_ensemble_success_rate_results(self, sr, label):
        self.db_save.db_save_success_rate(sr, label, self.settings["key_rank_report_interval"])



from aisy_sca.analysis.SingleModel import SingleModel
import time


class MultiModel:

    def __init__(self, settings, dataset, da_function, random_states=None):
        self.settings = settings
        self.dataset = dataset
        self.labels_key_guesses_attack_set = None
        self.labels_key_guesses_validation_set = None
        self.analysis_db_controls = None
        self.da_function = da_function
        self.custom_callbacks = []
        self.random_states = random_states

    def run(self, models):
        start = time.time()

        for model_index, model in models.items():
            single_model = SingleModel(self.settings, self.dataset, self.da_function, random_states=self.random_states)
            single_model.labels_key_guesses_attack_set = self.labels_key_guesses_attack_set
            single_model.labels_key_guesses_validation_set = self.labels_key_guesses_validation_set
            single_model.analysis_db_controls = self.analysis_db_controls
            single_model.run(model)
            self.custom_callbacks.append(single_model.get_custom_callbacks())

        self.analysis_db_controls.update_results_in_database(time.time() - start)



from aisy_sca.optimization.RandomGridSearch import RandomGridSearch
from aisy_sca.optimization.ProfilingAnalyzer import ProfilingAnalyzer
from aisy_sca.optimization.Ensemble import Ensemble
from aisy_sca.AnalysisDatabaseControls import *
from aisy_sca.LeakageModelControls import *


class HyperparameterSearchProcess:

    def __init__(self, dataset, settings, models, da_function, hp_combinations_reproducible, labels_key_guesses_attack_set=None,
                 labels_key_guesses_validation_set=None, loss_function=None, random_states=None):
        self.dataset = dataset
        self.settings = settings
        self.models = models
        self.da_function = da_function
        self.hp_combinations_reproducible = hp_combinations_reproducible
        self.custom_callbacks = None
        self.labels_key_guesses_attack_set = labels_key_guesses_attack_set
        self.labels_key_guesses_validation_set = labels_key_guesses_validation_set
        self.loss_function = loss_function
        self.random_states = random_states

    def run(self):

        """ Create analysis in database """
        analysis_db_controls = AnalysisDatabaseControls(self.settings)
        self.settings["analysis_id"] = analysis_db_controls.analysis_id
        analysis_db_controls.save_leakage_model_to_database()

        """ Generate list of labels for all key guesses for validation and attack sets """
        if self.labels_key_guesses_attack_set is None and self.labels_key_guesses_validation_set is None:
            lm_control = LeakageModelControls(self.settings)
            self.labels_key_guesses_attack_set, self.labels_key_guesses_validation_set = lm_control.compute_labels_key_guesses(self.dataset)

        """ Create random search process"""
        if self.settings["use_random_search"]:
            search_settings = self.settings["random_search"]
        else:
            search_settings = self.settings["grid_search"]

        if self.settings["use_profiling_analyzer"]:
            hp_search = ProfilingAnalyzer(self.settings, self.dataset, search_settings, self.labels_key_guesses_attack_set,
                                          self.labels_key_guesses_validation_set, analysis_db_controls, random_states=self.random_states)
            hp_search.loss_function = self.loss_function
            hp_search.run(self.hp_combinations_reproducible, da_function=self.da_function, grid_search=self.settings["use_grid_search"])
            self.custom_callbacks = hp_search.get_custom_callbacks()
        else:
            hp_search = RandomGridSearch(self.settings, self.dataset, search_settings, self.labels_key_guesses_attack_set,
                                         self.labels_key_guesses_validation_set, analysis_db_controls, random_states=self.random_states)
            hp_search.loss_function = self.loss_function
            hp_search.run(self.hp_combinations_reproducible, da_function=self.da_function, grid_search=self.settings["use_grid_search"])
            self.custom_callbacks = hp_search.get_custom_callbacks()

        """ Compute Ensembles"""
        if self.settings["use_ensemble"]:
            ensembles = Ensemble(self.settings, self.dataset)
            if self.settings["use_profiling_analyzer"]:
                for n_profiling_traces in self.settings["profiling_analyzer_steps"]:
                    ensembles.compute(hp_search.get_sk_attack_pa()[f"{n_profiling_traces}"], self.settings["ensemble"][0],
                                      self.labels_key_guesses_attack_set, analysis_db_controls, n_profiling=n_profiling_traces)
            else:
                ensembles.compute(hp_search.get_sk_attack(), self.settings["ensemble"][0], self.labels_key_guesses_attack_set,
                                  analysis_db_controls)
            if self.settings["use_early_stopping"]:
                for metric in self.settings["early_stopping"]["metrics"]:
                    if self.settings["use_profiling_analyzer"]:
                        for n_profiling_traces in self.settings["profiling_analyzer_steps"]:
                            ensembles.compute(hp_search.get_sk_attack_es()[metric][f"{n_profiling_traces}"],
                                              self.settings["ensemble"][0], self.labels_key_guesses_attack_set,
                                              analysis_db_controls, early_stopping_metric=metric, n_profiling=n_profiling_traces)
                    else:
                        ensembles.compute(hp_search.get_sk_attack_es()[metric], self.settings["ensemble"][0],
                                          self.labels_key_guesses_attack_set, analysis_db_controls, early_stopping_metric=metric)

    def get_custom_callbacks(self):
        return self.custom_callbacks
